import numpy as np

from pymdma.common.definitions import FeatureMetric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ..functional.knn import compute_nearest_neighbour_distances, compute_pairwise_distance


class ImprovedPrecision(FeatureMetric):
    """Improved Precision Metric for accessing fidelity of generative models.

    **Objective**: Fidelity

    Parameters
    ----------
    k : int, optional
        Number of nearest neighbors to consider in the hypersphere estimation. Defaults to 5.
    metric : str, optional, default="euclidean"
        The metric to use when calculating distance between instances.
        For the available metrics, see the documentation of `sklearn.metrics.pairwise_distances`.
    n_workers : int, optional
        Number of workers for computing pairwise distances. Defaults to 4.
    **kwargs
        Additional keyword arguments for compatiblilty.

    References
    ----------
    Kynkaanniemi et al., Improved Precision and Recall Metric for Assessing Generative Models (2019).
    https://arxiv.org/abs/1904.06991

    Code adapted from:
    improved-precision-and-recall-metric: Improved Precision and Recall Metric for Assessing Generative Models — Official TensorFlow Implementation.
    https://github.com/kynkaat/improved-precision-and-recall-metric

    Hypersphere estimation code was taken from:
    generative-evaluation-prdc, Reliable Fidelity and Diversity Metrics for Generative Models.
    https://github.com/clovaai/generative-evaluation-prdc

    Examples
    --------
    >>> improved_precision = ImprovedPrecision()
    >>> real_features = np.random.rand(100, 100)
    >>> fake_features = np.random.rand(100, 100)
    >>> result: MetricResult = improved_precision.compute(real_features, fake_features)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        k: int = 5,
        metric: str = "euclidean",
        n_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.k = k
        self.metric = metric
        self.n_workers = n_workers

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Compute the Improved Precision metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing the real features.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing the fake features.

        Notes
        -----
        Intermediate computations can be stored in the `context` dictionary of the `kwargs` parameter.
        Usefull when calculating multiple metrics that share the same intermediate computations.

        Returns
        -------
        result: MetricResult
            Dataset-level and instance-level results for the precision metric.
        """
        state = kwargs.get("context", {})
        if "real_nn_distances" not in state:
            state["real_nn_distances"] = compute_nearest_neighbour_distances(
                real_features,
                nearest_k=self.k,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        if "real_fake_distances" not in state:
            state["real_fake_distances"] = compute_pairwise_distance(
                real_features,
                fake_features,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        precision = (
            np.logical_or(
                (state["real_fake_distances"] < np.expand_dims(state["real_nn_distances"], axis=1)),
                np.isclose(state["real_fake_distances"], np.expand_dims(state["real_nn_distances"], axis=1)),
            )
            .any(axis=0)
            .astype(int)
        )

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": precision.mean()},
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "int", "value": precision.tolist()},
        )


class ImprovedRecall(FeatureMetric):
    """Improved Recall Metric for accessing diversity of generative models.

    **Objective**: Diversity

    Parameters
    ----------
    k : int, optional
        Number of nearest neighbors to consider in the hypersphere estimation. Defaults to 5.
    metric : str, optional, default="euclidean"
        The metric to use when calculating distance between instances.
        For the available metrics, see the documentation of `sklearn.metrics.pairwise_distances`.
    n_workers : int, optional
        Number of workers for computing pairwise distances. Defaults to 4.
    **kwargs
        Additional keyword arguments for compatiblilty.

    References
    ----------
    Kynkaanniemi et al., Improved Precision and Recall Metric for Assessing Generative Models (2019).
    https://arxiv.org/abs/1904.06991

    Code adapted from:
    improved-precision-and-recall-metric: Improved Precision and Recall Metric for Assessing Generative Models — Official TensorFlow Implementation.
    https://github.com/kynkaat/improved-precision-and-recall-metric

    Hypersphere estimation code was taken from:
    generative-evaluation-prdc, Reliable Fidelity and Diversity Metrics for Generative Models.
    https://github.com/clovaai/generative-evaluation-prdc

    Examples
    --------
    >>> improved_recall = ImprovedRecall()
    >>> real_features = np.random.rand(100, 100)
    >>> fake_features = np.random.rand(100, 100)
    >>> result: MetricResult = improved_recall.compute(real_features, fake_features)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        k: int = 5,
        metric: str = "euclidean",
        n_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.k = k
        self.metric = metric
        self.n_workers = n_workers

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Compute the Improved Recall metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing the real features.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing the fake features.

        Notes
        -----
        Intermediate computations can be stored in the `context` dictionary of the `kwargs` parameter.
        Usefull when calculating multiple metrics that share the same intermediate computations.

        Returns
        -------
        result: MetricResult
            Dataset-level and instance-level results for the recall metric.
        """
        state = kwargs.get("context", {})
        if "fake_nn_distances" not in state:
            state["fake_nn_distances"] = compute_nearest_neighbour_distances(
                fake_features,
                nearest_k=self.k,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        if "real_fake_distances" not in state:
            state["real_fake_distances"] = compute_pairwise_distance(
                real_features,
                fake_features,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        recall_mask = np.logical_or(
            state["real_fake_distances"] < np.expand_dims(state["fake_nn_distances"], axis=0),
            np.isclose(state["real_fake_distances"], np.expand_dims(state["fake_nn_distances"], axis=0)),
        )
        recall = recall_mask.any(axis=1).astype(int)

        # matrix with (R, F) shape -> .any() -> matrix with (F,) shape
        # an array that indicates for each F sample how many real samples are within its manifold
        recall_counts = recall_mask.sum(axis=0)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": recall.mean()},
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "int", "value": recall_counts.tolist()},
        )


class Density(FeatureMetric):
    """Density Metric for accessing fidelity of the generated samples. Unlike
    Improved Precision, it is robust towards outliers in the real/reference
    data.

    **Objective**: Fidelity

    Parameters
    ----------
    k : int, optional
        Number of nearest neighbors to consider in the hypersphere estimation. Defaults to 5.
    metric : str, optional, default="euclidean"
        The metric to use when calculating distance between instances.
        For the available metrics, see the documentation of `sklearn.metrics.pairwise_distances`.
    n_workers : int, optional
        Number of workers for computing pairwise distances. Defaults to 4.
    **kwargs
        Additional keyword arguments for compatibility.

    References
    ----------
    Naeem et al., Reliable Fidelity and Diversity Metrics for Generative Models (2020).
    https://arxiv.org/abs/2002.09797

    Code was adapted from:
    generative-evaluation-prdc, Reliable Fidelity and Diversity Metrics for Generative Models.
    https://github.com/clovaai/generative-evaluation-prdc

    Examples
    --------
    >>> density = Density()
    >>> real_features = np.random.rand(100, 100)
    >>> fake_features = np.random.rand(100, 100)
    >>> result: MetricResult = density.compute(real_features, fake_features)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        k: int = 5,
        metric: str = "euclidean",
        n_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.k = k
        self.metric = metric
        self.n_workers = n_workers

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Compute the Density metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing the real features.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing the fake features.

        Notes
        -----
        Intermediate computations can be stored in the `context` dictionary of the `kwargs` parameter.
        Usefull when calculating multiple metrics that share the same intermediate computations.

        Returns
        -------
        result: MetricResult
            Dataset-level and instance-level results for the density metric.
        """
        state = kwargs.get("context", {})
        if "real_nn_distances" not in state:
            state["real_nn_distances"] = compute_nearest_neighbour_distances(
                real_features,
                nearest_k=self.k,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        if "real_fake_distances" not in state:
            state["real_fake_distances"] = compute_pairwise_distance(
                real_features,
                fake_features,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        density = np.logical_or(
            (state["real_fake_distances"] < np.expand_dims(state["real_nn_distances"], axis=1)),
            np.isclose(state["real_fake_distances"], np.expand_dims(state["real_nn_distances"], axis=1)),
        )
        density = (1.0 / float(self.k)) * density.sum(axis=0)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": density.mean()},
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": density.tolist()},
        )


class Coverage(FeatureMetric):
    """Coverage Metric for accessing diversity of the generated samples. Unlike
    Improved Recall, it is robust towards outliers in the real/reference data.

    **Objective**: Diversity

    Parameters
    ----------
    k : int, optional
        Number of nearest neighbors to consider in the hypersphere estimation. Defaults to 5.
    metric : str, optional, default="euclidean"
        The metric to use when calculating distance between instances.
        For the available metrics, see the documentation of `sklearn.metrics.pairwise_distances`.
    n_workers : int, optional
        Number of workers for computing pairwise distances. Defaults to 4.
    **kwargs
        Additional keyword arguments for compatibility.

    References
    ----------
    Naeem et al., Reliable Fidelity and Diversity Metrics for Generative Models (2020).
    https://arxiv.org/abs/2002.09797

    Code was adapted from:
    generative-evaluation-prdc, Reliable Fidelity and Diversity Metrics for Generative Models.
    https://github.com/clovaai/generative-evaluation-prdc

    Examples
    --------
    >>> coverage = Coverage()
    >>> real_features = np.random.rand(100, 100)
    >>> fake_features = np.random.rand(100, 100)
    >>> result: MetricResult = coverage.compute(real_features, fake_features)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        k: int = 5,
        metric: str = "euclidean",
        n_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.k = k
        self.metric = metric
        self.n_workers = n_workers

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Compute the Coverage metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing the real features.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing the fake features.

        Notes
        -----
        Intermediate computations can be stored in the `context` dictionary of the `kwargs` parameter.
        Usefull when calculating multiple metrics that share the same intermediate computations.

        Returns
        -------
        result: MetricResult
            Dataset-level and instance-level results for the coverage metric.
        """
        state = kwargs.get("context", {})
        if "real_nn_distances" not in state:
            state["real_nn_distances"] = compute_nearest_neighbour_distances(
                real_features,
                nearest_k=self.k,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        if "real_fake_distances" not in state:
            state["real_fake_distances"] = compute_pairwise_distance(
                real_features,
                fake_features,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        coverage = np.logical_or(
            state["real_fake_distances"].min(axis=1) < state["real_nn_distances"],
            np.isclose(state["real_fake_distances"].min(axis=1), state["real_nn_distances"]),
        )

        # matrix with (R, F) shape -> .any() -> matrix with (F,) shape
        # an array that indicates for each F in how many real manifolds it is contained in
        coverage_counts = np.logical_or(
            state["real_fake_distances"] < np.expand_dims(state["real_nn_distances"], axis=1),
            np.isclose(state["real_fake_distances"], np.expand_dims(state["real_nn_distances"], axis=1)),
        ).sum(axis=0)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": coverage.mean()},
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "int", "value": coverage_counts.tolist()},
        )


class Authenticity(FeatureMetric):
    """Authenticity Metric for assessing the authenticity of the generated
    samples. A synthetic sample is considered authentic if it is signficantly
    distinct from any real sample.

    **Objective**: Privacy

    Parameters
    ----------
    metric : str, optional, default="euclidean"
        The metric to use when calculating distance between instances.
        For the available metrics, see the documentation of `sklearn.metrics.pairwise_distances`.
    **kwargs
        Additional keyword arguments for compatibility.

    Notes
    -----
    The authenticity metric is computed by checking if any fake sample is closer to a real sample than the real sample is to any other real sample.

    References
    ----------
    Alaa et al., How Faithful Is Your Synthetic Data? Sample-Level Metrics for Evaluating and Auditing Generative Models. (2022)
    https://doi.org/10.48550/arXiv.2102.08921.

    Hypersphere estimation code was adapted from:
    generative-evaluation-prdc, Reliable Fidelity and Diversity Metrics for Generative Models.
    https://github.com/clovaai/generative-evaluation-prdc
    
    Examples
    --------
    >>> authenticity = Authenticity()
    >>> real_features = np.random.rand(100, 100)
    >>> fake_features = np.random.rand(100, 100)
    >>> result: MetricResult = authenticity.compute(real_features, fake_features)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.PRIVACY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        metric: str = "euclidean",
        n_workers: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.metric = metric
        self.n_workers = n_workers

    def compute(self, real_features: np.ndarray, fake_features: np.ndarray, **kwargs) -> MetricResult:
        """Compute the Authenticity metric.

        Parameters
        ----------
        real_features : np.ndarray
            Array of shape (n_samples, n_features) containing the real features.
        fake_features : np.ndarray
            Array of shape (n_samples, n_features) containing the fake features.

        Notes
        -----
        Intermediate computations can be stored in the `context` dictionary of the `kwargs` parameter.
        Usefull when calculating multiple metrics that share the same intermediate computations.

        Returns
        -------
        result: MetricResult
            Dataset-level and instance-level results for the authenticity metric
        """
        state = kwargs.get("context", {})

        if "real_fake_distances" not in state:
            state["real_fake_distances"] = compute_pairwise_distance(
                real_features,
                fake_features,
                metric=self.metric,
                n_workers=self.n_workers,
            )

        # compute distance to closest real samples
        state["real_closest_real_distances"] = compute_nearest_neighbour_distances(
            real_features,
            nearest_k=1,
            metric=self.metric,
            n_workers=self.n_workers,
        )

        # check if any fake sample is closer to Ri than Ri is to any other Rj
        authenticity = np.logical_or(
            state["real_fake_distances"] < np.expand_dims(state["real_closest_real_distances"], axis=1),
            np.isclose(state["real_fake_distances"], np.expand_dims(state["real_closest_real_distances"], axis=1)),
        )

        # mask of the values that are considered authentic in the fake dataset
        authenticity_mask = ~authenticity.any(axis=0)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": authenticity_mask.mean()},
            instance_level={
                "dtype": OutputsTypes.ARRAY,
                "subtype": "int",
                "value": authenticity_mask.astype(int).tolist(),
            },
        )


__all__ = ["ImprovedPrecision", "ImprovedRecall", "Density", "Coverage", "Authenticity"]



from pymdma.image.measures.input_val.data.dom import DOM
from pymdma.image.measures.input_val.data.no_reference import (
    BRISQUE,
    CLIPIQA,
    EME,
    Brightness,
    Colorfulness,
    ExposureBrightness,
    Tenengrad,
    TenengradRelative,
)
from pymdma.image.measures.input_val.data.psnr import PSNR
from pymdma.image.measures.input_val.data.ssim import MSSSIM, SSIM

__all__ = [
    "BRISQUE",
    "Tenengrad",
    "TenengradRelative",
    "EME",
    "DOM",
    "ExposureBrightness",
    "Brightness",
    "Colorfulness",
    "CLIPIQA",
    "PSNR",
    "SSIM",
    "MSSSIM",
]



import torch
import torchvision.models as tvmodels
from PIL import Image
from piq.feature_extractors import InceptionV3

from .extractor import BaseExtractor


class InceptionFID(BaseExtractor):
    def __init__(self):
        super().__init__(
            input_size=(299, 299),
            interpolation=Image.Resampling.BILINEAR,
        )

        self.extractor = InceptionV3(normalize_input=False)

    def forward(self, x):
        N = len(x)
        x = self.extractor(x)
        return x[0].view(N, -1)


class VGGExtractor(BaseExtractor):
    def __init__(self, model_name: str) -> None:
        super().__init__(
            input_size=(224, 224),
            interpolation=Image.Resampling.BILINEAR,
        )

        weights = tvmodels.vgg.__dict__[f"{model_name.upper()}_Weights"].DEFAULT
        self.extractor = tvmodels.vgg.__dict__[model_name](weights=weights)
        # features from the last fully connected layer
        self.extractor.classifier = self.extractor.classifier[:-2]

    def forward(self, x):
        return self.extractor(x)


class ViTExtractor(BaseExtractor):
    def __init__(self, model_name: str) -> None:
        super().__init__(
            input_size=(224, 224),
            interpolation=Image.Resampling.BILINEAR,
        )

        weights = tvmodels.vision_transformer.__dict__[f"{model_name.upper().replace('VIT', 'ViT')}_Weights"].DEFAULT
        self.extractor = tvmodels.vision_transformer.__dict__[model_name](weights=weights)

    def forward(self, x):
        # Reshape and permute the input tensor
        x = self.extractor._process_input(x)
        n = x.shape[0]

        # Expand the class token to the full batch
        batch_class_token = self.extractor.class_token.expand(n, -1, -1)
        x = torch.cat([batch_class_token, x], dim=1)
        x = self.extractor.encoder(x)
        # # Classifier "token" as used by standard language architectures
        x = x[:, 0]
        return x


class DinoExtractor(BaseExtractor):
    def __init__(self, model_name, input_size: tuple[int, int] = (224, 224)) -> None:
        super().__init__(
            input_size=input_size,
            interpolation=Image.Resampling.BICUBIC,
        )

        # get model from the hub without classifier heads
        self.extractor = torch.hub.load(
            f"facebookresearch/{model_name.split('_')[0]}:main",
            model_name,
            pretrained=True,
        )

    def forward(self, batch):
        return self.extractor(batch)



from typing import List, Literal

import numpy as np
from fastdtw import fastdtw

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType


class DTW(Metric):
    """Computes the Dynamic Time Warping (DTW) distance between two sets of
    time-series signals, evaluating the similarity between corresponding
    channels in the target and reference signals. The DTW distance is computed
    by comparing each target signal with every reference signal, with lower DTW
    values indicating greater similarity. For each signal pair, the DTW
    distance is calculated across all channels, with the mean of the distances
    being taken across all channels and instances. This process yields both
    instance-level and dataset-level DTW metrics.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility with the Metric framework.

    References
    ----------
    Salvador, S., & Chan, P., FastDTW: Toward Accurate Dynamic Time (2004).
    https://dl.acm.org/doi/10.5555/1367985.1367993

    Examples
    --------
    >>> dtw = DTW()
    >>> reference_sigs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> target_imgs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = dtw.compute(reference_sigs, target_sigs)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def compute(
        self,
        reference_sigs: List[np.ndarray],
        target_sigs: List[np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Computes Dinamic Time Wrapping.

        Parameters
        ----------
        reference_sigs: (N, L, C) ndarray
            Signals to use as reference.
            List of arrays representing a signal of shape (L, C).
        target_sigs : (N, L, C) ndarray
            Signals compare with reference.
            List of arrays representing a signal of shape (L, C).

        Returns
        -------
        result : MetricResult
            Instance-level dtw.
            Dataset-level dtw.
        """
        instance_dtw = []
        for targ in target_sigs:
            for ref in reference_sigs:
                dtw_values_by_chan = []

                # Compute the dtw for each channel
                for targ_channel, ref_channel in zip(targ.T, ref.T):
                    channel_dtw, _ = fastdtw(targ_channel, ref_channel)

                    dtw_values_by_chan.append(channel_dtw)

                # Compute mean dtw across channels
                mean_dtw = np.mean(dtw_values_by_chan)

            # Correlations by signal
            instance_dtw.append(mean_dtw)

        # Average correlation across signals
        final_dtw = np.mean(instance_dtw)

        return MetricResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": instance_dtw},
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": final_dtw},
        )


class CrossCorrelation(Metric):
    """Computes the Cross-Correlation between two sets of signals.

    This function calculates the cross-correlation to analyze the relationship between
    corresponding channels in the target and reference signals. The computation is performed
    for each signal in the target set against every signal in the reference set, using a
    specified overlap mode. The computed cross-correlation for each channel can be summarized
    using one of two reduction methods: 'mean'and 'max'.

    For each signal pair, the function calculates the cross-correlation values for each channel
    using the specified reduction method. It then computes the mean of these values across all
    channels to provide an instance-level metric and averages these results across all signal
    pairs to obtain the dataset-level metric.

    **Objective**: Fidelity, Diversity

    Parameters
    ----------
    mode : {'full', 'same', 'valid'}, optional
        Defines how the cross-correlation is computed. Default is 'full'.
        - 'full': Computes the convolution at every point of overlap, producing
          an output of size (N + M - 1). Boundary effects may be present.
        - 'same': Produces an output of length max(M, N), centered on the signals.
        - 'valid': Produces an output of length max(M, N) - min(M, N) + 1,
          considering only complete overlaps.
    reduction : {'mean', 'max'}, optional
        Determines how the cross-correlation is summarized for each channel. Default is 'max'.
        - 'mean': The average of the cross-correlation values for the channel.
        - 'max': The maximum cross-correlation value for the channel.
    **kwargs : dict, optional
        Additional keyword arguments for customization.

    References
    ----------
    Proakis and Manolakis, Digital Signal Processing: Principles, Algorithms, and Applications (1996).
    https://dl.acm.org/doi/10.5555/227373

    Examples
    --------
    >>> cc = CrossCorrelation()
    >>> reference_sigs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> target_sigs = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = cc.compute(reference_sigs, target_sigs)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = [EvaluationLevel.INSTANCE, EvaluationLevel.DATASET]
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = -np.inf
    max_value: float = np.inf

    def __init__(
        self,
        mode: Literal["full", "same", "valid"] = "full",
        reduction: Literal["mean", "max"] = "max",
        **kwargs,
    ):
        super().__init__(**kwargs)
        assert mode in ["full", "same", "valid"], f"Unsupported mode for Cross Correlation: {mode}"
        self.mode = mode
        assert reduction in ["mean", "max"], f"Unsupported criteria for relative tenengrad: {reduction}"
        self.reduction = reduction

    def compute(
        self,
        reference_sigs: List[np.ndarray],
        target_sigs: List[np.ndarray],
        **kwargs,
    ) -> MetricResult:
        """Computes Dinamic Time Wrapping.

        Parameters
        ----------
        reference_sigs: (N, L1, C) ndarray
            Signals to use as reference.
            List of arrays representing a signal of shape (L1, C).
        target_sigs :(N, L2, C) ndarray
            Signals compare with reference.
            List of arrays representing a signal of shape (L2, C).

        Returns
        -------
        result : MetricResult
            Instance-level maximum cross-correlation.
            Dataset-level maximum cross-correlation.
        """
        instance_cross_corr = []

        for targ in target_sigs:
            for ref in reference_sigs:
                reduct_corr_values_by_chan = []

                # Compute the cross-correlation for each channel with different reductions
                for targ_channel, ref_channel in zip(targ.T, ref.T):
                    cross_corr = np.correlate(targ_channel, ref_channel, mode=self.mode)

                    if self.reduction == "max":
                        max_corr_idx = np.argmax(cross_corr)
                        max_corr_value = cross_corr[max_corr_idx]
                        reduct_corr_values_by_chan.append(max_corr_value)

                    else:
                        mean_corr_value = np.mean(cross_corr)
                        reduct_corr_values_by_chan.append(mean_corr_value)

                # Compute the mean of the correlation across channels
                cross_corr = np.mean(reduct_corr_values_by_chan)

            # Correlations by signal
            instance_cross_corr.append(cross_corr)

        # Average correlation across signals
        final_cross_corr = np.mean(instance_cross_corr)

        return MetricResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": instance_cross_corr},
            dataset_level={"dtype": OutputsTypes.NUMERIC, "subtype": "float", "value": final_cross_corr},
        )


__all__ = ["DTW", "CrossCorrelation"]



import inspect
from importlib import import_module
from pathlib import Path
from typing import Dict, List, Optional, Union

from loguru import logger

from pymdma.common.definitions import InputLayer, Metric
from pymdma.config import package_base_dir
from pymdma.constants import (
    METRICS_PACKAGE_NAME,
    DataModalities,
    EvaluationLevel,
    InputCategories,
    MetricGroup,
    ReferenceType,
    SyntheticCategories,
    ValidationDomain,
)


def _get_constant_values(class_name):
    # Get all attributes of the class
    attributes = [
        attr for attr in dir(class_name) if not callable(getattr(class_name, attr)) and not attr.startswith("__")
    ]
    # Filter out only the constants and get their values
    constant_values = [getattr(class_name, attr) for attr in attributes if not callable(getattr(class_name, attr))]
    return constant_values


def _check_class_attribute(source, target):
    if isinstance(target, list):
        return any(source == t for t in target)
    return source == target


def select_modality_input_layer(
    data_modality: DataModalities,
    validation_domain: ValidationDomain,
    reference_type: ReferenceType,
    target_data: Path,
    reference_data: Optional[Path] = None,
    batch_size: int = 20,
    output_dir: Optional[Path] = None,
    annotation_file: Optional[Path] = None,
    device: str = "cpu",
) -> InputLayer:
    """Selects the input layer based on the data modality.

    Parameters
    ----------
    data_modality : str
        The data modality to be used (time_series, image, tabular, text)
    validation_domain : str
        The validation type to be used (input_val or synthesis_val)
    reference_type : str
        The reference type (dataset-wise or instance-wise)
    target_data : Path
        The path to the target data
    reference_data : Path, optional
        The path to the reference data
    batch_size : int, optional
        The batch size to be used (default=20)
    output_dir : Path, optional
        The path to the output directory (default=None)
    annotation_file : Path, optional
        The path to the annotation file (default=None)
    device : str, optional
        The device to be used (default=cpu)

    Returns
    -------
    input_layer : InputLayer
        The input layer
    """
    if data_modality == "image":
        from pymdma.image.input_layer import ImageInputLayer

        # allow for repeat reference in  full reference input validation
        repeat_reference = validation_domain == ValidationDomain.INPUT and reference_type == ReferenceType.INSTANCE
        return ImageInputLayer(
            validation_domain,
            reference_type,
            target_data,
            reference_data,
            batch_size=batch_size,
            repeat_reference=repeat_reference,
            device=device,
            features_cache=output_dir,
            annot_file=annotation_file,
        )
    elif data_modality == "time_series":
        from pymdma.time_series.input_layer import TimeSeriesInputLayer

        return TimeSeriesInputLayer(
            validation_domain,
            reference_type,
            target_data,
            reference_data,
            batch_size=batch_size,
        )
    elif data_modality == "tabular":
        from pymdma.tabular.input_layer import TabularInputLayer

        return TabularInputLayer(
            validation_domain,
            reference_type,
            target_data,
            reference_data,
        )
    elif data_modality == "text":
        from pymdma.text.input_layer import TextInputLayer

        return TextInputLayer(
            validation_domain,
            reference_type,
            target_data,
            reference_data,
            batch_size=batch_size,
            device=device,
        )
    else:
        raise ValueError("Unknown data modality")


def import_module_or_folder(validation_root: Path, group_name: str) -> List[str]:
    """Import a module or a folder of modules."""
    if not validation_root.exists():
        raise FileNotFoundError(f"Validation root not found: {validation_root}")

    group_root = validation_root / group_name
    # module is a folder
    if group_root.is_dir():
        return [
            f"{group_name}.{module.stem}"
            for module in group_root.iterdir()
            if module.is_file() and module.suffix == ".py" and module.stem != "__init__"
        ]

    # module is a file
    if group_root.with_suffix(".py").is_file():
        return [group_name]
    # module is not found
    return []


def select_metric_classes(
    data_modality: DataModalities,
    validation_domain: ValidationDomain,
    metric_categorys: Optional[List[Union[SyntheticCategories, InputCategories]]] = None,
    ignore_missing: bool = False,
) -> Dict[str, List[Metric]]:
    """Helper function for importing classes from the metrics package."""
    # script names
    module_root = ".".join((data_modality, METRICS_PACKAGE_NAME, validation_domain))
    # fetch all modules when None # TODO review for current structure

    # fetch all metric groups in a module
    if metric_categorys is None:
        metric_categorys = [path.stem for path in Path(f"{package_base_dir}/{module_root.replace('.', '/')}").iterdir()]
        metric_categorys = [group for group in metric_categorys if "__" not in group]

    group_classes: Dict[str, List[Metric]] = {}
    for metric_category in metric_categorys:
        group_root = module_root.replace(".", "/")
        group_root = Path(f"{package_base_dir}/{group_root}")

        group_modules = import_module_or_folder(group_root, metric_category)

        if len(group_modules) < 1 and not ignore_missing:
            logger.error(
                f"No modules found for {group_root} and the provided metric groups: {metric_categorys}. Check if the provided parameters are correct.",
            )
            return group_classes

        for module in group_modules:
            module_path = f"{package_base_dir.split('/')[-1]}.{module_root}.{module}"
            try:
                module = import_module(module_path)
            except ModuleNotFoundError as e:
                logger.error(f"Error loading module: {module_path}. Detail: {e}")
                return {}

            # Get classes names
            classes_in_module = [
                obj
                for _name, obj in inspect.getmembers(
                    module,
                    lambda member: inspect.isclass(member)
                    and issubclass(member, Metric)
                    and not member.__name__ in {"Metric", "FeatureMetric"},
                )
            ]
            group_classes.setdefault(metric_category, []).extend(classes_in_module)
    return group_classes


def select_metric_functions(
    data_modality: DataModalities,
    validation_domain: ValidationDomain,
    reference_type: ReferenceType,
    evaluation_level: Optional[EvaluationLevel] = None,
    metric_category: Optional[Union[SyntheticCategories, InputCategories]] = None,
    metric_groups: Optional[List[MetricGroup]] = None,
) -> Dict[str, List[Metric]]:
    """Helper function for selecting specific subset of measures.

    Parameters
    ----------
    data_modality : str
        The data modality to be used (time_series, image, tabular, text)
    validation_domain : str
        The validation type to be used (input_val or synthesis_val)
    reference_type : str
        The reference type (dataset-wise or instance-wise)
    evaluation_level : str, optional (default=None)
        The evaluation level (dataset-wise or instance-wise)
    metric_category : str, optional (default=None)
        The metric group to be used
    metric_groups : list, optional (default=None)
        The metric goals to be used

    Returns
    -------
    selected_functions : Dict[str, List[object]]
        The selected functions
    """

    if isinstance(metric_groups, str):
        metric_groups = [metric_groups]

    # get evaluation group class objects
    evaluation_group_classes = select_metric_classes(data_modality, validation_domain, metric_category)

    # class name
    if len(evaluation_group_classes) == 0:
        logger.error(
            f"No classes found for {data_modality}.{validation_domain} and the provided metric groups: {metric_category}. Check if the provided parameters are correct.",
        )
        return {}

    # filter functions based on reference, evaluation and metric_group
    selected_functions = {}
    for evaluation_group, metric_classes in evaluation_group_classes.items():
        # Filter out only the methods defined in the class and exclude those starting with "_"
        selected_functions.setdefault(evaluation_group, [])
        selected_names = set()
        for class_obj in metric_classes:
            # class_obj: Metric
            # discard classes with missing attributes
            if metric_groups is None or class_obj.metric_group in metric_groups:
                if (
                    evaluation_level is None or _check_class_attribute(evaluation_level, class_obj.evaluation_level)
                ) and _check_class_attribute(
                    reference_type,
                    class_obj.reference_type,
                ):
                    class_name = class_obj.__name__
                    selected_functions[evaluation_group] += [class_obj] if class_name not in selected_names else []
                    selected_names.add(class_name)
    return selected_functions


def select_specific_metric_functions(
    metric_names: List[str],
    data_modality: DataModalities,
    validation_domain: ValidationDomain,
    reference_type: Optional[ReferenceType] = None,
):
    """Helper function for selecting metrics by name."""
    # get evaluation group class objects
    evaluation_group_classes = select_metric_classes(data_modality, validation_domain)

    assert (
        len(evaluation_group_classes) > 0
    ), f"No groups found for {data_modality}.{validation_domain}. Check if the provided parameters are correct."

    # assert all(len(metric_fns) > 0 for metric_fns in evaluation_group_classes.values()), "One or more missing metrics. Check if the provided names are correct."

    metric_names = set(metric_names)

    selected_functions = {}
    for eval_group, group_classes in evaluation_group_classes.items():
        for class_obj in group_classes:
            class_name = class_obj.__name__
            if class_name in metric_names and (reference_type is None or reference_type == class_obj.reference_type):
                selected_functions.setdefault(eval_group, [])
                selected_functions[eval_group].append(class_obj)
    return selected_functions


def get_metrics_metadata(
    data_modalities: Optional[DataModalities] = None,
    validation_domains: Optional[ValidationDomain] = None,
    metric_categorys: Optional[Union[SyntheticCategories, InputCategories]] = None,
):
    """Get metadata for all metrics within the specified modalities, validation
    types and metric groups."""
    if data_modalities is None:
        data_modalities = _get_constant_values(DataModalities)
    elif isinstance(data_modalities, str):
        data_modalities = [data_modalities]

    if validation_domains is None:
        validation_domains = _get_constant_values(ValidationDomain)
    elif isinstance(validation_domains, str):
        validation_domains = [validation_domains]

    metrics = {}
    for data_modality in data_modalities:
        for validation_domain in validation_domains:
            evaluation_group_classes = select_metric_classes(
                data_modality,
                validation_domain,
                metric_categorys,
                ignore_missing=True,
            )
            # class name

            if len(evaluation_group_classes) == 0:
                continue

            # Get methods from class
            for evaluation_group, metric_classes in evaluation_group_classes.items():
                # Filter out only the methods defined in the class and exclude those starting with "_"
                selected_metrics = {
                    f"{data_modality}.{validation_domain}.{evaluation_group}.{method.__name__}": {
                        "name": method.__name__,
                        "data_modality": data_modality,
                        "validation_domain": validation_domain,
                        "metric_category": evaluation_group,
                        "description": method.__doc__,
                        **vars(method),
                    }
                    for method in metric_classes
                }

                metrics.update(selected_metrics)
    return metrics



import json
from pathlib import Path


def jsonl_files(file: Path, path_col: str = "image"):
    """Extracts the image paths from a jsonl dataset file."""
    with file.open("r") as f:
        return [Path(json.loads(line)[path_col]) for line in f]



# from typing import List, Optional

# import numpy as np

# from pymdma.common.definitions import Metric
# from pymdma.common.output import MetricResult
# from pymdma.constants import EvaluationLevel, OutputsTypes, ReferenceType, SyntheticFeatureMetrics

# from ...utils_syn import compute_utility_scores


# class Utility(Metric):
#     """Computes the utility evaluation scores for synthetic datasets. It
#     calculates the percentage change of performance scores presented as [1] RR
#     (Real vs Real), [2] SR (Synthetic vs Real), and [3] SRR (Synthetic and Real
#     vs Real).

#     Parameters
#     ----------
#     column_names : list, optional
#         The names of the columns to be considered for the utility calculation.
#     target_col : str, optional
#         The name of the target column to predict.
#     train_cols : list, optional
#         The list of columns to be used for training the model.

#     Returns
#     -------
#     MetricResult
#         A MetricResult object containing the utility scores.

#     Examples
#     --------
#     >>> # Example 1: Evaluating utility scores for a dataset
#     >>> import numpy as np
#     >>> real_data = np.concatenate((
#     ...     np.random.rand(100, 3),
#     ...     np.random.choice([0, 1], 100, p=[0.5, 0.5]).reshape(-1, 1)),
#     ...     axis=-1
#     ... )
#     >>> syn_data = np.concatenate((
#     ...     np.random.rand(100, 3),
#     ...     np.random.choice([0, 1], 100, p=[0.5, 0.5]).reshape(-1, 1)),
#     ...     axis=-1
#     ... )
#     >>> utility_score = Utility(column_names=['att1', 'att2', 'att3', 'target'], target_col='target')
#     >>> result: MetricResult = utility_score.compute(real_data, syn_data)
#     >>> dataset_level, _ = result.value # Output: utility scores for RR, SR, and SRR

#     >>> # Example 2: Using specific training columns
#     >>> utility_score_specific = Utility(column_names=['att1', 'att2', 'att3', 'target'],
#     ...                                    target_col='target', train_cols=['att1', 'att2'])
#     >>> result_specific: MetricResult = utility_score_specific.compute(real_data, syn_data)
#     >>> dataset_level, _ = result_specific.value # Output: utitility scores with specified training columns
#     """

#     reference_type = ReferenceType.DATASET
#     evaluation_level = EvaluationLevel.DATASET
#     metric_group = SyntheticFeatureMetrics.UTILITY

#     higher_is_better: bool = True
#     min_value: float = 0
#     max_value: float = 100.0

#     def __init__(
#         self,
#         column_names: Optional[List[str]] = None,
#         target_col: Optional[str] = None,
#         train_cols: Optional[List[str]] = None,
#         **kwargs,
#     ):
#         """Initializes the Utility metric to evaluate the utility scores for
#         synthetic datasets.

#         Parameters
#         ----------
#         column_names : list, optional
#             The names of the columns to be considered for the utility calculation.
#         target_col : str, optional
#             The name of the target column to predict.
#         train_cols : list, optional
#             The list of columns to be used for training the model.
#         **kwargs : dict
#             Additional keyword arguments passed to the parent class.
#         """

#         super().__init__(**kwargs)

#         # column list
#         self.column_names = column_names

#         # target column
#         self.target_col = target_col

#         # train columns
#         self.train_cols = train_cols

#     def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
#         """Computes the utility evaluation scores comparing synthetic and real
#         datasets.

#         Parameters
#         ----------
#         real_data : np.ndarray
#             The target dataset for evaluation.
#         syn_data : np.ndarray
#             The synthetic dataset to evaluate.
#         **kwargs : dict
#             Additional keyword arguments for computation.

#         Returns
#         -------
#         MetricResult
#             A MetricResult object containing the utility scores.
#         """

#         kind = ["rr", "sr", "srr"]

#         # define columns
#         if isinstance(self.column_names, (list, np.ndarray)):
#             cols_ = self.column_names
#         else:
#             cols_ = [f"att{x}" for x in range(real_data.shape[-1])]

#         # target column
#         targ_ = cols_[-1] if self.target_col is None else self.target_col

#         # train columns
#         train_ = cols_[:-1] if self.train_cols is None else self.train_cols

#         # perform utility using ML
#         score_map = compute_utility_scores(
#             real_data=real_data,
#             syn_data=syn_data,
#             cols=cols_,
#             target_col=targ_,
#             train_cols=train_,
#             kind=kind,
#         )

#         # get utility metric names
#         # metric_names = list(score_map["rr"].columns)

#         # build final score dict
#         score_final_d = {
#             # 'metric': metric_names,
#             "RR": score_map["rr"].round(3).loc[0].to_list(),
#             "SR": score_map["sr"].round(3).loc[0].to_list(),
#             "SRR": score_map["srr"].round(3).loc[0].to_list(),
#         }

#         return MetricResult(
#             dataset_level={
#                 "dtype": OutputsTypes.KEY_ARRAY,
#                 "subtype": "float",
#                 "value": score_final_d,
#             },
#         )


# __all__ = ["Utility"]



import os
import platform
from os import environ, getenv
from os.path import dirname, join
from pathlib import Path

import dotenv

# Project root
project_dir = os.getcwd()

# Load the environment variables from the `.env` file, overriding any system environment variables
env_path = join(project_dir, ".env")
dotenv.load_dotenv(env_path, override=True)

# Load secrets from the `.secrets` file, overriding any system environment variables
secrets_path = join(project_dir, ".secrets")
dotenv.load_dotenv(secrets_path, override=True)

# Some common paths
_reports_dir = Path(str(getenv("DIR_REPORTS")))
report_dir = join(project_dir, _reports_dir)

_figures_dir = Path(str(getenv("DIR_FIGURES")))
figures_dir = join(project_dir, _figures_dir)

_models_dir = Path(str(getenv("DIR_MODELS")))
models_dir = join(project_dir, _models_dir)

_notebook_dir = Path(str(getenv("DIR_NOTEBOOKS")))
notebook_dir = join(project_dir, _notebook_dir)

_data_dir = Path(str(getenv("DIR_DATA")))
data_dir = join(project_dir, _data_dir)

_data_raw_dir = Path(str(getenv("DIR_DATA_RAW")))
data_raw_dir = join(project_dir, _data_raw_dir)

_data_interim_dir = Path(str(getenv("DIR_DATA_INTERIM")))
data_interim_dir = join(project_dir, _data_interim_dir)

_data_processed_dir = Path(str(getenv("DIR_DATA_PROCESSED")))
data_processed_dir = join(project_dir, _data_processed_dir)

_data_test_dir = Path(str(getenv("DIR_DATA_TEST")))
data_test_dir = join(project_dir, _data_test_dir)

_package_dir = Path(str(getenv("PACKAGE_NAME")))
package_base_dir = join(project_dir, _package_dir)

# CUDA Enable
ENABLE_CUDA = True
if not ENABLE_CUDA:
    environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    environ["CUDA_VISIBLE_DEVICES"] = "-1"
    environ["USE_CPU"] = "1"

# Hydra
environ["HYDRA_FULL_ERROR"] = "1"

# log to mlflow
LOG_TO_MLFLOW = False

_IS_WINDOWS = platform.system() == "Windows"

# EOF



from typing import List, Union


def set_property(key: str, value: Union[List[str], str]):
    """Create a decorator that sets a specific property of the decorated
    function to a given value.

    Parameters
    ----------
    key : str
        The name of the property to be set.
    value : Any
        The value to be assigned to the specified property key.

    Returns
    -------
    decorator : function
        A decorator function that sets the specified property of a decorated function.
    """

    def decorator(func):
        setattr(func, key, value)
        return func

    return decorator


def inherit_annotations(cls):
    annotations = {}
    for parent in cls.__bases__:
        annotations.update(getattr(parent, "__annotations__", {}))
    cls.__annotations__.update(annotations)
    return cls



def dispersion_ratio(func, x_feat_1, x_feat_2, y_feat_1, y_feat_2):
    """Calculates the ratio of the distance between fake samples and the
    distance between real samples, using the metric passed in the argument
    "func".

    The dispersion ratio is a measure of how well the generator model
    is able to replicate the dispersion or spread of features present
    in the real samples.

    ratio = 1 --> distance between fake samples is equal to the
                  distance between real samples.(ideal scenario)
    ratio > 1 --> distance between fake samples is greater then
                  the distance between real samples.
    ratio < 1 --> distance between fake samples is smaller then
                  the distance between real samples.

    Parameters
    ----------
    func : callable
        A distance metric or function that takes two arrays of features
        as input and returns a scalar distance. This function is used to
        compute distances between samples.
    x_feat_1 : array-like of shape [n_samples/2, n_features]
        2D array with features of the original samples.
    x_feat_2 : array-like of shape [n_samples/2, n_features]
        2D array with features of the fake samples.
    y_feat_1 : array-like of shape [n_samples/2, n_features]
        2D array with features of the original samples.
    y_feat_2 : array-like of shape [n_samples/2, n_features]
        2D array with features of the fake samples.

    Returns
    -------
    dispersion_ratio : float
        The dispersion ratio calculated with the specified distance metric.
    """
    rr = func(x_feat_1, x_feat_2)
    ff = func(y_feat_1, y_feat_2)

    return ff / rr


def distance_ratio(func, x_feat_1, x_feat_2, y_feat_1, y_feat_2):
    """Calculates the ratio of the distance between real and fake samples and
    the distance of between real samples, using the metric passed in the
    argument "func".

    The distance ratio is a measure of how well the generator model
    is able to replicate the features from the real samples.

    ratio = 1 --> distance between real and fake samples is equal
                  to the distance between real samples.(ideal scenario)
    ratio > 1 --> distance between real and fake samples is greater
                  then the distance between real samples.
    ratio < 1 --> distance between real and fake samples is smaller
                  then the distance between real samples.

    Parameters
    ----------
    func : callable
        A distance metric or function that takes two arrays of features
        as input and returns a scalar distance. This function is used to
        compute distances between samples.
    x_feat_1 : array-like of shape [n_samples/2, n_features]
        2D array with features of the original samples.
    x_feat_2 : array-like of shape [n_samples/2, n_features]
        2D array with features of the fake samples.
    y_feat_1 : array-like of shape [n_samples/2, n_features]
        2D array with features of the original samples.
    y_feat_2 : array-like of shape [n_samples/2, n_features]
        2D array with features of the fake samples.

    Returns
    -------
    distance_ratio : float
        The distance ratio computed with the specified distance metric.
    """
    rr = func(x_feat_1, x_feat_2)
    fr = (func(x_feat_1, y_feat_2) + func(x_feat_2, y_feat_1)) / 2

    return fr / rr



from typing import Optional

from loguru import logger
from sklearn.preprocessing import MaxAbsScaler, MinMaxScaler, RobustScaler, StandardScaler

from ..embeddings import GenericNorm


class StandardNorm(GenericNorm):
    """Standard normalization class using StandardScaler to scale data to have
    a mean of 0 and a standard deviation of 1. Inherits from the GenericNorm
    base class.

    Example
    -------
    >>> scaler = StandardNorm()
    >>> X = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
    >>> scaler.fit_transform(X)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            self.scaler_obj = StandardScaler(**kwargs)
        except Exception as e:
            logger.debug(f"Error: {e}")
            self.scaler_obj = StandardScaler()

    def fit(self, X, y=None):
        self.scaler_obj.fit(X, y)
        return self

    def transform(self, X):
        return self.scaler_obj.transform(X)

    def fit_transform(self, X, y=None):
        return self.scaler_obj.fit_transform(X, y)

    def inverse_transform(self, X):
        return self.scaler_obj.inverse_transform(X)


class RobustNorm(GenericNorm):
    """Robust normalization class using RobustScaler to scale data by removing
    the median and scaling according to the interquartile range. Inherits from
    GenericNorm.

    Example
    -------
    >>> scaler = RobustNorm()
    >>> X = [[1, 2], [2, 4], [4, 8], [5, 10]]
    >>> scaler.fit_transform(X)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            self.scaler_obj = RobustScaler(**kwargs)
        except Exception as e:
            logger.debug(f"Error: {e}")
            self.scaler_obj = RobustScaler()

    def fit(self, X, y=None):
        self.scaler_obj.fit(X, y)
        return self

    def transform(self, X):
        return self.scaler_obj.transform(X)

    def fit_transform(self, X, y=None):
        return self.scaler_obj.fit_transform(X, y)

    def inverse_transform(self, X):
        return self.scaler_obj.inverse_transform(X)


class MinMaxNorm(GenericNorm):
    """Min-max normalization class using MinMaxScaler to scale data to a given
    range (default between 0 and 1). Inherits from GenericNorm.

    Example
    -------
    >>> scaler = MinMaxNorm()
    >>> X = [[1, 2], [2, 3], [3, 4], [4, 5]]
    >>> scaler.fit_transform(X)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            self.scaler_obj = MinMaxScaler(**kwargs)
        except Exception as e:
            logger.debug(f"Error: {e}")
            self.scaler_obj = MinMaxScaler()

    def fit(self, X, y=None):
        self.scaler_obj.fit(X, y)
        return self

    def transform(self, X):
        return self.scaler_obj.transform(X)

    def fit_transform(self, X, y=None):
        return self.scaler_obj.fit_transform(X, y)

    def inverse_transform(self, X):
        return self.scaler_obj.inverse_transform(X)


class MaxAbsNorm(GenericNorm):
    """Max-abs normalization class using MaxAbsScaler to scale data based on
    the absolute maximum value for each feature. Inherits from GenericNorm.

    Example
    -------
    >>> scaler = MaxAbsNorm()
    >>> X = [[1, -1], [2, -2], [3, -3], [4, -4]]
    >>> scaler.fit_transform(X)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            self.scaler_obj = MaxAbsScaler(**kwargs)
        except Exception as e:
            logger.debug(f"Error: {e}")
            self.scaler_obj = MaxAbsScaler()

    def fit(self, X, y=None):
        self.scaler_obj.fit(X, y)
        return self

    def transform(self, X):
        return self.scaler_obj.transform(X)

    def fit_transform(self, X, y=None):
        return self.scaler_obj.fit_transform(X, y)

    def inverse_transform(self, X):
        return self.scaler_obj.inverse_transform(X)


class GlobalNorm:
    """A wrapper class that applies the specified normalization method based on
    the provided normalization type.

    Parameters
    ----------
    n_type : Optional[str]
        The type of normalization method to use.
        Must be one of: "umap", "pca", "lle", "isomap", "spacy", "bert", "identity".

    Example
    -------
    >>> scaler = GlobalNorm(n_type='minmax')
    >>> X = [[1, 2], [2, 3], [3, 4], [4, 5]]
    >>> scaler.fit_transform(X)
    """

    N_MAP = {
        "maxabs": MaxAbsNorm,
        "minmax": MinMaxNorm,
        "standard": StandardNorm,
        "robust": RobustNorm,
        "identity": GenericNorm,
    }

    def __init__(self, n_type: Optional[str] = "standard", **kwargs):
        if isinstance(n_type, str) and n_type in self.N_MAP:
            # tag
            norm_tag = n_type.lower()

            # norm class
            cls_ = self.N_MAP.get(norm_tag, GenericNorm)

            # norm instance
            self.scaler_obj = cls_(**kwargs)
        else:
            self.scaler_obj = GenericNorm()

    def fit(self, X, y=None):
        self.scaler_obj.fit(X, y)
        return self

    def transform(self, X):
        return self.scaler_obj.transform(X)

    def fit_transform(self, X, y=None):
        return self.scaler_obj.fit_transform(X, y)

    def inverse_transform(self, X):
        return self.scaler_obj.inverse_transform(X)



from typing import List, Tuple

import numpy as np
import pandas as pd
from loguru import logger
from scipy.stats import gaussian_kde
from sklearn.model_selection import train_test_split
from sklearn.neighbors import NearestNeighbors

from ..data.utils import is_categorical, is_numeric
from ..models.classification import ClassificationML
from ..models.regression import RegressionML

device = "cpu"  # matrices are too big for gpu
feature_metrics = {"categorical": "hamming", "numerical": "euclidean"}


def _get_cdf(data, nbins: int = 50):
    # histogram
    count, _ = np.histogram(data, bins=nbins)

    # probability distribution function
    pdf = count / sum(count)

    # cumulative distribution function
    cdf = np.cumsum(pdf)

    return cdf


def _get_change_perc(df1: pd.DataFrame, df2: pd.DataFrame):
    try:
        perc = (df1 - df2).divide(df2 + 1e-20) * 100
    except TypeError as e:
        logger.debug(f"Error:\n{e}")
        perc = None
    return perc


def _get_nn_model(data: np.ndarray, distance_type: str = "euclidean"):
    """Find nearest neighbors of test in train with first categoric_slice-many
    variables being categorical."""
    # get nn model
    nearest_neighbor_model = NearestNeighbors(
        metric=distance_type,
        algorithm="ball_tree",
        n_jobs=None,
    )

    # fit data
    nearest_neighbor_model.fit(data)

    return nearest_neighbor_model


def _get_nn_distances(
    tgt_emb: np.ndarray,
    syn_emb: np.ndarray,
    distance_type: dict = "euclidean",
    size: int = None,
) -> Tuple[np.ndarray]:
    # checkpoint
    assert tgt_emb.shape[1] == syn_emb.shape[1], "Train and Syn have mismatched columns"

    # split into tgt_train, tgt_query, and syn_query
    if size is None:
        tgt_size, syn_size = len(tgt_emb), len(syn_emb)
    else:
        tgt_size, syn_size = size, size

    # train and query from target
    tgt_query = tgt_emb[-int(tgt_size) :]

    # syn_train is not needed
    # if sample_size = synthetic_size, syn_query is all syn dataset
    syn_query = syn_emb[-int(syn_size) :]

    # training model
    nn_model = _get_nn_model(tgt_query, distance_type)

    # get nearest neighbors
    # target
    tgt_query_nn, _ = nn_model.kneighbors(tgt_query, n_neighbors=3)
    tgt_query_nn = tgt_query_nn[:, 1:]  # except the closest (itself)

    # synthetic
    syn_query_nn, _ = nn_model.kneighbors(syn_query, n_neighbors=2)

    # calculating DCR NNDR
    query_dict = {"tgt": tgt_query_nn, "syn": syn_query_nn}

    # compute privacy scores
    dcr, nndr = {}, {}
    for label, query in query_dict.items():
        # closest neighbor
        aux_dcr = query[:, 0]

        # normalized closest neighbor distances
        aux_nndr = aux_dcr / (query[:, 1] + 1e-10)

        # assign
        dcr[label] = aux_dcr
        nndr[label] = aux_nndr

    return dcr, nndr


def _get_gaussian_kde(data: np.ndarray, bins: np.ndarray):
    # get unique values
    unq = np.unique(data)

    # check condition
    if len(unq) > 1:
        # sample from distribution
        pdf_data = gaussian_kde(
            data,
        ).pdf(bins)

        # normalize
        pdf_data /= sum(pdf_data)

    else:
        # single value bin index
        arg_idx = np.argmin(abs(bins - unq[0]))

        # gen default distribution
        pdf_data = np.zeros_like(bins, dtype=float)

        # assign single value bin
        pdf_data[arg_idx] = 1.0

    return pdf_data


def _get_nn_pdf(
    tgt_dist: np.ndarray,
    syn_dist: np.ndarray,
) -> Tuple[np.ndarray]:

    # get distributions bins
    t_min, t_max = min(tgt_dist), max(tgt_dist)
    s_min, s_max = min(syn_dist), max(syn_dist)

    bins = np.linspace(
        min([t_min, s_min]),
        max([t_max, s_max]),
        300,
    )

    # get distributions
    # tgt pdf dists
    pdf_tgt = _get_gaussian_kde(tgt_dist, bins)

    # syn pdf dists
    pdf_syn = _get_gaussian_kde(syn_dist, bins)

    return pdf_tgt, pdf_syn, bins


def _get_pp_metrics(
    tgt_pdf: np.ndarray,
    syn_pdf: np.ndarray,
    bins: np.ndarray,
    low_perc: int = 10,
    high_perc: int = 90,
    tag: str = "dist",
):
    # overlap
    overlap = np.minimum(tgt_pdf, syn_pdf).sum()

    # percentiles real
    tgt_cumsum = np.cumsum(tgt_pdf)
    p_low = np.where(tgt_cumsum >= low_perc / 100)[0][0]
    p_high = np.where(tgt_cumsum >= high_perc / 100)[0][0]

    # percentiles synthetic
    syn_cumsum = np.cumsum(syn_pdf)
    p_low_s = np.where(syn_cumsum >= low_perc / 100)[0][0]
    p_high_s = np.where(syn_cumsum >= high_perc / 100)[0][0]

    # synthetic samples within percentiles
    syn_leak = syn_pdf[:p_low].sum()
    syn_nleak = syn_pdf[p_high:].sum()
    syn_fidel = 1 - syn_leak - syn_nleak
    syn_div = (p_high_s - p_low_s) / (max([p_high, p_high_s]) - min([p_low, p_low_s]))

    # centroid displacement
    tgt_ctd = np.dot(tgt_pdf, bins)
    syn_ctd = np.dot(syn_pdf, bins)
    ctd_disp = (syn_ctd - tgt_ctd) / (max(bins) - min(bins))
    ctd_disp = (ctd_disp + 1) / 2  # min-max norm (-1/1 to 0/1)

    # metrics dict
    metric_d = {
        "dist": tag,
        "%_overlap": round(overlap * 100, 1),
        "%_leak": round(syn_leak * 100, 1),
        "%_nfidel": round(syn_nleak * 100, 1),
        "%_fidel": round(syn_fidel * 100, 1),
        "%_divers": round(syn_div * 100, 1),
        "%_ctd_disp": round(ctd_disp * 100, 1),
        "bin_low": bins[p_low],
        "bin_high": bins[p_high],
    }

    # result string
    metric_s = " | ".join([f"{m.split('_')[-1]}:{val}%" for m, val in metric_d.items() if "syn" in m])

    # add to dict
    metric_d["prompt"] = metric_s

    return metric_d


def assign_utility_experiment(
    train_tgt: pd.DataFrame,
    train_syn: pd.DataFrame,
    test_tgt: pd.DataFrame,
    target_col: str = "",
    kind: str = "rr",
    task_class: callable = ClassificationML,
    **kwargs,
):
    # assign experiment
    if kind == "rr":
        train_, test_ = train_tgt, test_tgt
    elif kind == "sr":
        train_, test_ = train_syn, test_tgt
    elif kind == "srr":
        train_, test_ = pd.concat((train_tgt, train_syn), axis=0), test_tgt
    else:
        train_, test_ = None, None

    exp = task_class(train_, test_, target_col)

    return exp


def compute_utility_scores(
    real_data: np.ndarray,
    syn_data: np.ndarray,
    cols: list,
    target_col: str,
    train_cols: list,
    kind: List = ["rr", "sr", "srr"],
    seed: int = 84,
    **kwargs,
):

    # columns (target and train)
    job_cols = train_cols + [target_col]

    # real data
    X_real_total = pd.DataFrame(real_data, columns=cols)[job_cols]

    # synthetic data
    X_syn_total = pd.DataFrame(syn_data, columns=cols)[job_cols]

    # detect target variable type (continuous, categorical)
    is_categ = is_categorical(
        np.concatenate((X_real_total[target_col], X_syn_total[target_col])),
    )

    # split real data for train/test
    X_real_train, X_real_test, _, _ = train_test_split(
        X_real_total,
        X_real_total,
        random_state=seed,
        stratify=(X_real_total[target_col] if is_categ else None),
    )

    # reset indices
    X_train = X_real_train.reset_index(drop=True)
    X_test = X_real_test.reset_index(drop=True)

    # select the task type
    task_class = ClassificationML if is_categ else RegressionML

    # utility evaluation strategy
    score_map = {}
    for k in kind:
        # Utility experiment
        utility_exp = assign_utility_experiment(
            train_tgt=X_train,
            train_syn=X_syn_total,
            test_tgt=X_test,
            target_col=target_col,
            kind=k,
            task_class=task_class,
        )

        # run
        if is_categ:
            utility_exp.prepare_run_(
                normalize=True,
                fix_imbalance=True,
            )
        else:
            utility_exp.prepare_run_(
                normalize=True,
            )

        # compare_models
        run_ = utility_exp.do_run_(
            cross_validation=False,
        )

        # pick best model
        if is_categ:
            # get dataframe results
            df_scores = utility_exp.exp.pull().sort_values("Model").iloc[:, 1:-3]

            # classification (maximize score)
            df_scores = df_scores[df_scores["F1"] == df_scores["F1"].max()].iloc[:1]
        else:
            # get dataframe results
            df_scores = utility_exp.exp.pull().sort_values("Model").iloc[:, [1, 2, 3]]

            # regression (minimize score)
            df_scores = df_scores[df_scores["MAE"] == df_scores["MAE"].min()].iloc[:1]

        logger.debug(df_scores)

        # reset indices
        df_scores.reset_index(drop=True, inplace=True)

        # assign to mapper
        score_map[k] = df_scores

    return score_map


def _get_kl_divergence(real_pdf: np.ndarray, syn_pdf: np.ndarray):
    """Compute the Kullback-Leibler (KL) divergence between two probability
    distributions.

    Parameters:
    - real_pdf, syn_pdf: 1-D series representing the two probability distributions.

    Returns:
    - kl_score: KL divergence score.
    """

    eps = 1e-10

    # compute KL divergence
    kl_score = np.ma.masked_invalid(
        np.where(real_pdf != 0, real_pdf * np.log(real_pdf / (syn_pdf + eps)), 0),
    ).sum()

    return np.maximum(kl_score, 0)


def _get_js_divergence(real_pdf: np.ndarray, syn_pdf: np.ndarray):
    """Compute the Jensen-Shannon (JS) divergence between two probability
    distributions.

    Parameters:
    - real_pdf, syn_pdf: 1-D series representing the two probability distributions.

    Returns:
    - js_score: JS divergence score.
    """

    eps = 1e-10

    # Calculate midpoint distribution
    m = 0.5 * (real_pdf + syn_pdf)

    # Compute JS divergence
    kl_tgt = np.ma.masked_invalid(real_pdf * np.log2(real_pdf / (m + eps))).sum()
    kl_syn = np.ma.masked_invalid(syn_pdf * np.log2(syn_pdf / (m + eps))).sum()

    js_score = 0.5 * (kl_tgt + kl_syn)

    return np.maximum(js_score, 0)


def _get_ks_similarity(real_data: np.ndarray, syn_data: np.ndarray):
    """Compute attribute similarity for continuous attributes using Kolmogorov-
    Smirnov statistics.

    Parameters
    ----------
    real_col : np.ndarray
        The target dataset attribute.
    syn_col : np.ndarray
        The synthetic dataset attribute.

    Returns
    ----------
    float
        KS score
    """
    # define bin range
    _min = np.concatenate([real_data, syn_data]).min()  # min
    _max = np.concatenate([real_data, syn_data]).max()  # max

    bins = np.linspace(_min, _max, 100)  # histogram bins

    # original cdf
    real_cdf = _get_cdf(real_data, nbins=bins)

    # synthetic cdf
    syn_cdf = _get_cdf(syn_data, nbins=bins)

    # ks statistic
    ks_score = 1 - np.max(real_cdf - syn_cdf)

    return ks_score


def _get_tv_similarity(real_data: np.ndarray, syn_data: np.ndarray) -> float:
    """Compute attribute similarity for categorical attributes.

    Parameters
    ----------
    real_col : np.ndarray
        The target dataset attribute.
    syn_col : np.ndarray
        The synthetic dataset attribute.

    Returns
    -------
    float
        TV similarity score
    """

    # original cdf
    real_unq, real_counts = np.unique(real_data, return_counts=True)
    real_map = {unq: count / sum(real_counts) for unq, count in zip(real_unq, real_counts)}

    # synthetic cdf
    syn_unq, syn_counts = np.unique(syn_data, return_counts=True)
    syn_map = {unq: count / sum(syn_counts) for unq, count in zip(syn_unq, syn_counts)}

    # tv statistics
    scores = [abs(real_map.get(unq) - syn_map.get(unq, 0)) for unq in real_unq]

    # tv score
    tv_score = 1 - (np.sum(scores) / len(real_unq))

    return tv_score



from pathlib import Path
from typing import Callable, List, Union

import numpy as np
import torch
import tsfel
from torch import nn
from torch.utils.data import DataLoader

from pymdma.common.definitions import EmbedderInterface

from ..data.simple_dataset import _read_sig_file


class BaseTSExtractor(nn.Module, EmbedderInterface):
    default: str = "tsfel"
    extractor: Union[nn.Module, Callable] = None

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__()
        pass

    @torch.no_grad()
    def extract_features_from_files(self, files: List[Path], fs: int, dims: List, batch_size: int = 4):
        """Extract features from a list of image files.

        Args:
        files: (List[Path])
            list of paths to signal files
        batch_size: int
            batch size for feature extraction. Defaults to 4.

        Returns
        -------
        np.ndarray
            array of features
        """
        if batch_size > len(files):
            batch_size = len(files)

        assert len(files) > 0, "No files to extract features from."

        n_batches = len(files) // batch_size
        batch_sizes = [batch_size for _ in range(n_batches)]
        if len(files) % batch_size != 0:
            batch_sizes.append(len(files) % batch_size)

        act_array = []
        start, end = 0, 0
        for bsize in batch_sizes:
            end = start + bsize
            signals = [_read_sig_file(f) for f in files[start:end]]
            batch = self(signals, fs, dims)
            act_array.append(batch)
            start += bsize
        return np.concatenate(act_array, axis=0)

    @torch.no_grad()
    def _extract_features_dataloader(self, dataloader: DataLoader, fs: int, dims: List):
        """Use selected approach to extract features from all signals in the
        dataloader.

        Parameters
        ----------
        dataloader: DataLoader
            signals dataloader
        fs: int
            sampling frequency
        dims: List
             list containing the names of each signal dimension or channel.
             For example, in the context of ECG data, this would be a list of the names of each ECG lead.

        Returns
        -------
        Tuple[np.ndarray, List, List]
            Array of features, list of signal labels and list of signal IDs.
        """
        act_array = []
        labels_array = []
        ids_array = []

        for batch, labels, signal_ids in dataloader:
            batch_feat = self(batch, fs, dims)
            act_array.append(batch_feat)
            labels_array.extend(labels)
            ids_array.extend(signal_ids)

        features = np.concatenate(act_array, axis=0)
        return features, labels_array, ids_array



import numpy as np

from pymdma.common.definitions import FeatureMetric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ...utils_syn import _get_nn_distances, _get_nn_pdf, _get_pp_metrics


class NNDRPrivacy(FeatureMetric):
    """Compute the NNDR (Nearest Neighbor Distance Ratio) privacy score. A more
    negative score indicates better privacy assurance, while a less negative
    score suggests higher risk.

    **Objective**: Authenticity

    Parameters
    ----------
    distance_type : str, optional, default="euclidean"
        The distance metric to use for calculating distances between embeddings.
        Default is "euclidean".
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Liu et al., Scaling while privacy preserving: A comprehensive synthetic tabular data generation and evaluation in learning analytics (2024).
    https://doi.org/10.1145/3636555.3636921

    Returns
    -------
    MetricResult
        A MetricResult object containing the DCR and NNDR privacy scores.

    Examples
    --------
    >>> import numpy as np
    >>> real_embeddings = np.random.rand(100, 10)  # 100 samples, 10 features
    >>> synthetic_embeddings = np.random.rand(100, 10)
    >>> nndr_privacy_score = NNDRPrivacy(distance_type="euclidean")
    >>> result: MetricResult = nndr_privacy_score.compute(real_embeddings, synthetic_embeddings)
    >>> dataset_level, _ = result.value  # Output: privacy metrics including DCR and NNDR scores
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.PRIVACY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 100.0

    def __init__(
        self,
        distance_type: str = "euclidean",
        **kwargs,
    ):
        super().__init__(**kwargs)

        # distance type with which metrics are calculated
        self.distance_type = distance_type

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Computes the NNDR privacy score comparing synthetic and real dataset
        embeddings.

        Parameters
        ----------
        real_data : np.ndarray
            The target dataset embeddings.
        syn_data : np.ndarray
            The synthetic dataset embeddings.
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the privacy metrics including DCR and NNDR scores.
        """

        # get sample-sample distances
        _, nndr_map = _get_nn_distances(real_data, syn_data, self.distance_type)

        # get NNDR distributions
        tgt_pdf, syn_pdf, bins = _get_nn_pdf(
            nndr_map["tgt"],
            nndr_map["syn"],
        )

        # get NNDR scores
        metric_d = _get_pp_metrics(
            tgt_pdf,
            syn_pdf,
            bins,
            tag="NNDR",
            low_perc=25,
            high_perc=75,
        )

        # privacy metrics
        priv_d = {
            "privacy": 100 - metric_d.get("%_leak"),
            "level": metric_d.get("%_ctd_disp"),
        }

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_VAL,
                "subtype": "float",
                "value": priv_d,
            },
        )


class DCRPrivacy(FeatureMetric):
    """Compute the DCR (Distance to Closest Record) privacy score. A more
    negative score indicates better privacy assurance, while a less negative
    score suggests higher risk.

    **Objective**: Authenticity

    Parameters
    ----------
    distance_type : str, optional, default="euclidean"
        The distance metric to use for calculating distances between embeddings.
        Default is "euclidean". Available from: `link <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.distance_metrics.html#sklearn.metrics.pairwise.distance_metrics>`_
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Liu et al., Scaling while privacy preserving: A comprehensive synthetic tabular data generation and evaluation in learning analytics (2024).
    https://doi.org/10.1145/3636555.3636921

    Returns
    -------
    MetricResult
        A MetricResult object containing the DCR and NNDR privacy scores.

    Examples
    --------
    >>> import numpy as np
    >>> real_emb = np.random.rand(100, 10)  # 100 samples, 10 features
    >>> synthetic_emb = np.random.rand(100, 10)
    >>> dcr_privacy_score = DCRPrivacy(distance_type="euclidean")
    >>> result: MetricResult = dcr_privacy_score.compute(real_emb, synthetic_emb)
    >>> dataset_level, _ = result.value  # Output: privacy metrics including DCR and NNDR scores
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.PRIVACY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 100.0

    def __init__(
        self,
        distance_type: str = "euclidean",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.distance_type = distance_type

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Computes the DCR privacy score comparing synthetic and real dataset
        embeddings.

        Parameters
        ----------
        real_data : np.ndarray
            The target dataset embeddings.
        syn_data : np.ndarray
            The synthetic dataset embeddings.
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the privacy metrics including DCR and NNDR scores.
        """

        # get sample-sample distances
        dcr_map, _ = _get_nn_distances(real_data, syn_data, self.distance_type)

        # get DCR distributions
        tgt_pdf, syn_pdf, bins = _get_nn_pdf(
            dcr_map["tgt"],
            dcr_map["syn"],
        )

        # get DCR scores
        metric_d = _get_pp_metrics(
            tgt_pdf,
            syn_pdf,
            bins,
            tag="DCR",
            low_perc=25,
            high_perc=75,
        )

        # privacy metrics
        priv_d = {
            "privacy": 100 - metric_d.get("%_leak"),
            "level": metric_d.get("%_ctd_disp"),
        }

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_VAL,
                "subtype": "float",
                "value": priv_d,
            },
        )


__all__ = ["NNDRPrivacy", "DCRPrivacy"]



from typing import Dict, List, Optional, Tuple, Union

from loguru import logger
from pycocotools.coco import COCO

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import AnnotationType, EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

_SUPPORTED_ANNOT_TYPES = {
    "segmentation": AnnotationType.MASK,
    "bbox": AnnotationType.BBOX,
    "keypoints": AnnotationType.KEYPOINTS,
    "category_id": AnnotationType.LABEL,
}


def infer_annotation_types(dataset: Dict[str, any]) -> List[AnnotationType]:
    """Infer annotation types from the dataset."""
    # infer annotation types from dataset
    annotation_types = set()
    for annotation in dataset["annotations"]:
        annotation_types.update(annotation.keys() & _SUPPORTED_ANNOT_TYPES.keys())
    annotation_types = [_SUPPORTED_ANNOT_TYPES[annot_type] for annot_type in annotation_types]
    # logger.info(f"Annotation types found: {annotation_types}")
    return annotation_types


def get_anns_from_img_id(dataset: Dict[str, any], img_id: int) -> List[int]:
    """Get all annotations from a single image."""
    return [ann for ann in dataset["annotations"] if "image_id" in ann and ann["image_id"] == img_id]


def get_anns_from_cat_id(dataset: Dict[str, any], cat_id: int) -> List[int]:
    """Get all annotations with a specific category id."""
    return [ann for ann in dataset["annotations"] if "category_id" in ann and ann["category_id"] == cat_id]


class DatasetCompletness(Metric):
    """Evalute the completeness of the COCO dataset.

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    References
    ----------
    Lin et al., Microsoft COCO: Common Objects in Context (2014).
    https://arxiv.org/abs/1405.0312

    Examples
    --------
    >>> import json
    >>> dataset_completness = DatasetCompletness()
    >>> with open("annotations.json", "r") as f:
    ...     dataset = json.load(f)
    >>> result: MetricResult = dataset_completness.compute(dataset)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = [EvaluationLevel.DATASET, EvaluationLevel.INSTANCE]
    metric_group = MetricGroup.VALIDITY
    annotation_type = [AnnotationType.LABEL, AnnotationType.BBOX, AnnotationType.MASK, AnnotationType.KEYPOINTS]

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def _completness_labels(self, dataset: Dict[str, any]) -> List[int]:
        """Checks whether all images have at least one category assigned and if
        it is different from 0.

        In COCO format format, a categoy_id = 0 means no label.
        """
        img_id_no_label = []

        category_ids = {cat["id"] for cat in dataset["categories"]}

        for image in dataset["images"]:
            annotations_for_image = get_anns_from_img_id(dataset, image["id"])
            # annotations_for_image = dataset.loadAnns(dataset.getAnnIds(imgIds=image_id))
            if len(annotations_for_image) == 0:
                # logger.warning(f"No annotations found for image {image_id}")
                img_id_no_label.append(image["id"])

            for annot in annotations_for_image:
                if "category_id" not in annot:
                    # logger.warning(f"Category_id not found for annotation {annot['id']}")
                    img_id_no_label.append(annot["id"])
                    continue
                if annot["category_id"] == 0 or annot["category_id"] not in category_ids:
                    img_id_no_label.append(annot["id"])
                    # logger.warning(f"Invalid label for image {image_id}")
        return img_id_no_label

    def _completness_annots(self, dataset: Dict[str, any]) -> List[int]:
        """Checks whether all images have at least one annotation for each
        annotation type type (bbox, mask, or keypoints).

        Returns
        -------
        value - list where the first element is a bool indicating success (True) or not (False) of the metric and the second element is a list representing the images ids where the metric failed (if verified).
        """

        annotation_types = {"bbox", "segmentation", "keypoints"}
        invalid_annot_ids = []
        for annot in dataset["annotations"]:
            if not annot.keys() & annotation_types:
                # logger.warning(f"Annotation {annot['id']} does not have any annotation.")
                invalid_annot_ids.append(annot["id"])
                continue
        return invalid_annot_ids

    def _coco_parsing(self, dataset: Dict[str, any]) -> Union[str, None]:
        try:
            coco = COCO(None)
            coco.dataset = dataset
            coco.createIndex()
        except Exception as e:
            logger.error(f"Error parsing COCO file: {e}")
            return f"Non standard COCO annotation, review incorrect fields: {str(e)}"
        return None

    def compute(
        self,
        dataset: Dict[str, any],
        **kwargs,
    ) -> MetricResult:
        """Compute the dataset completeness metric.

        Parameters
        ----------
        dataset : dict
            Dictionary with COCO dataset format.

        Returns
        -------
        result : MetricResult
            - Dataset-level metric result with the following keys:
                - "img_no_annot" - list of image ids with no annotations.
                - "annot_missing_fields" - list of annotation ids with missing fields.
            - Errors - list of errors found during the evaluation.
        """

        img_missing_labels = self._completness_labels(dataset)
        annot_missing_fields = self._completness_annots(dataset)

        coco_parsing_error = self._coco_parsing(dataset)

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_ARRAY,
                "subtype": "int",
                "value": {
                    "img_no_annot": img_missing_labels,
                    "annot_missing_fields": annot_missing_fields,
                },
            },
            errors=[coco_parsing_error] if coco_parsing_error is not None else None,
        )


class AnnotationCorrectness(Metric):
    """Evalute the annotation correctness of the COCO dataset.

    Parameters
    ----------
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    References
    ----------
    Lin et al., Microsoft COCO: Common Objects in Context (2014).
    https://arxiv.org/abs/1405.0312

    Examples
    --------
    >>> import json
    >>> ann_correct = AnnotationCorrectness()
    >>> with open("annotations.json", "r") as f:
    ...     dataset = json.load(f)
    >>> result: MetricResult = ann_correct.compute(dataset)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = [EvaluationLevel.DATASET, EvaluationLevel.INSTANCE]
    metric_group = MetricGroup.VALIDITY
    annotation_type = [AnnotationType.LABEL, AnnotationType.BBOX, AnnotationType.MASK, AnnotationType.KEYPOINTS]

    def __init__(
        self,
        annotation_types: List[AnnotationType] = None,
        num_keypoints: Optional[int] = None,
        mask_area_range: Optional[Tuple[int, int]] = None,
        valid_label_names: Optional[List[str]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.annotation_types = annotation_types
        self.num_keypoints = num_keypoints

        assert (
            mask_area_range is None or type(mask_area_range) in {tuple, list} and len(mask_area_range) == 2
        ), "mask_area_range must be a tuple or list with two elements."
        self.min_mask_area = mask_area_range[0] if mask_area_range else None
        self.max_mask_area = mask_area_range[1] if mask_area_range else None

        self.valid_label_names = valid_label_names

    def _correctness_bbox(self, dataset: Dict[str, any]) -> List[int]:
        """Checks whether the bounding box coordinates are valid (i.e., have 4
        coords)."""
        invalid_bbox_ann_ids = []
        for annotation in dataset["annotations"]:
            if "bbox" in annotation and len(annotation["bbox"]) != 4:
                invalid_bbox_ann_ids.append(annotation["id"])
                # logger.warning(f"Invalid bbox for annotation {annotation['id']}: {annotation['bbox']}")
        return invalid_bbox_ann_ids

    def _correctness_keypoints(self, dataset: Dict[str, any]) -> List[int]:
        """Checks whether the number of keypoints is valid for each image."""
        annot_ids_incomplete_keypoints = []

        if not self.num_keypoints:
            logger.warning("Number of keypoints not provided. Skipping keypoint evaluation.")
            return []

        for annotation in dataset["annotations"]:
            if "keypoints" not in annotation:
                continue
            kp = annotation["keypoints"]
            if len(kp) != self.num_keypoints:
                annot_ids_incomplete_keypoints.append(annotation["id"])
                # logger.warning(f"Invalid number of keypoints for annotation {annotation['id']}: {kp}")

    def _correctness_mask(self, dataset: Dict[str, any]) -> Dict[str, List[int]]:
        """Checks whether the mask area, counts, size, and bbox are valid for
        each mask annotation."""
        area_mask_bounds = []
        missing_area = []
        missing_counts = []
        missing_size = []
        missing_bbox = []

        for annotation in dataset["annotations"]:
            if "segmentation" not in annotation:
                continue
            if "area" not in annotation:
                missing_area.append(annotation["id"])
            else:
                area = annotation["area"]
                if (self.min_mask_area is not None and area < self.min_mask_area) or (
                    self.max_mask_area is not None and area > self.max_mask_area
                ):
                    area_mask_bounds.append(annotation["id"])
                    # logger.warning(f"Invalid area for annotation {annotation['id']}: {area}")
            if "counts" not in annotation["segmentation"]:
                # logger.warning(f"Missing counts for annotation {annotation['id']}")
                missing_counts.append(annotation["id"])
            if "size" not in annotation["segmentation"]:
                # logger.warning(f"Missing size for annotation {annotation['id']}")
                missing_size.append(annotation["id"])
            if "bbox" not in annotation:
                # logger.warning(f"Missing bbox for annotation {annotation['id']}")
                missing_bbox.append(annotation["id"])
        return {
            "annots_mask_oob": area_mask_bounds,
            "annots_missing_area": missing_area,
            "annots_missing_counts": missing_counts,
            "annots_missing_size": missing_size,
            "annots_missing_bbox": missing_bbox,
        }

    def _correctness_labels(self, dataset: Dict[str, any]) -> List[int]:
        """Checks whether the categories contained in the annotations file are
        valid and if annotation corresponds to an invalid category id."""
        invalid_labels = []
        annots_ids_invalid_labels = []

        for cat_values in dataset["categories"]:
            cat_id = cat_values["id"]
            if self.valid_label_names and cat_values["name"] not in self.valid_label_names:
                invalid_labels.append(cat_id)
                # logger.warning(f"Invalid category found: {cat_values['name']}")
                invalid_cat_anns = get_anns_from_cat_id(dataset, cat_id)
                annots_ids_invalid_labels.extend([ann["id"] for ann in invalid_cat_anns])
        return invalid_labels, annots_ids_invalid_labels

    def compute(
        self,
        dataset: Dict[str, any],
        **kwargs,
    ) -> MetricResult:
        """Compute the annotation completeness metric.

        Parameters
        ----------
        dataset : dict
            Dictionary with COCO dataset format.

        Returns
        -------
        result : MetricResult
            - Dataset-level metric result with the following keys:
                - "invalid_labels" - list of categories with invalid label names.
                - "annots_invalid_label" - list of annotation ids with invalid label names.
                - "annots_invalid_bbox" - list of annotation ids with invalid bounding boxes.
                - "annots_invalid_kp" - list of annotation ids with invalid keypoints.
                - "annots_mask_oob" - list of annotation ids with masks out of bounds.
                - "annots_missing_area" - list of annotation ids with missing area.
                - "annots_missing_counts" - list of annotation ids with missing counts.
                - "annots_missing_size" - list of annotation ids with missing size.
                - "annots_missing_bbox" - list of annotation ids with missing bounding boxes.
        """

        annotation_errors = {}
        self.annotation_types = (
            infer_annotation_types(dataset) if self.annotation_types is None else self.annotation_types
        )
        for annot_type in self.annotation_types:
            if annot_type == AnnotationType.LABEL:
                invalid_labels, annots_ids_invalid_labels = self._correctness_labels(dataset)
                annotation_errors["invalid_labels"] = invalid_labels
                annotation_errors["annots_invalid_label"] = annots_ids_invalid_labels
            if annot_type == AnnotationType.BBOX:
                invalid_bbox_ann_ids = self._correctness_bbox(dataset)
                annotation_errors["annots_invalid_bbox"] = invalid_bbox_ann_ids
            if annot_type == AnnotationType.KEYPOINTS:
                annot_ids_incomplete_keypoints = self._correctness_keypoints(dataset)
                annotation_errors["annots_invalid_kp"] = annot_ids_incomplete_keypoints
            if annot_type == AnnotationType.MASK:
                mask_errors = self._correctness_mask(dataset)
                annotation_errors.update(mask_errors)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.KEY_ARRAY, "subtype": "int", "value": annotation_errors},
        )


class AnnotationUniqueness(Metric):
    """Evalute the annotation uniqueness of the COCO dataset.

    Parameters
    ----------
    annotation_types : list
        List of annotation types to evaluate. Default is ["bbox", "segmentation", "keypoints"].
    **kwargs : dict, optional
        Additional keyword arguments for compatibility (unused).

    Notes
    -----
    This metric checks if there are duplicated annotations for each annotation type (bbox, mask, or keypoints) for the same image and if they have different categories assigned.
    This metric should only be used if the goal of your dataset is to have unique annotations for each image.

    References
    ----------
    Lin et al., Microsoft COCO: Common Objects in Context (2014).
    https://arxiv.org/abs/1405.0312

    Examples
    --------
    >>> import json
    >>> ann_unique = AnnotationUniqueness()
    >>> with open("annotations.json", "r") as f:
    ...     dataset = json.load(f)
    >>> result: MetricResult = ann_unique.compute(dataset)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = [EvaluationLevel.DATASET, EvaluationLevel.INSTANCE]
    metric_group = MetricGroup.VALIDITY
    annotation_type = [AnnotationType.BBOX, AnnotationType.MASK, AnnotationType.KEYPOINTS]

    def __init__(
        self,
        annotation_types: List[AnnotationType] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.annotation_types = annotation_types
        self.annotation_types = (
            {"bbox", "segmentation", "keypoints"} if self.annotation_types is None else self.annotation_types
        )

    def compute(
        self,
        dataset: Dict[str, any],
        **kwargs,
    ) -> MetricResult:
        """Compute the annotation completeness metric.

        Parameters
        ----------
        dataset : dict
            Dictionary with COCO dataset format.

        Returns
        -------
        result : MetricResult
            Dataset-level metric results with a dictionary that maps from image id to a list of repeated annotations
        """
        imgs_ids_repeated_annot = {}
        # image_id -> category_id -> annotation_types
        image_annot_type_map = {}
        for ann in dataset["annotations"]:
            if "image_id" not in ann or "category_id" not in ann:
                continue
            img_id = str(ann["image_id"])
            annot_cat = ann["category_id"]
            annot_types = list(ann.keys() & self.annotation_types)
            image_annot_type_map.setdefault(img_id, dict())
            image_annot_type_map[img_id].setdefault(annot_cat, set())

            # check if annotation type already exists for the image
            if annot_cat in image_annot_type_map[img_id]:
                if any(annot_type in image_annot_type_map[img_id][annot_cat] for annot_type in annot_types):
                    imgs_ids_repeated_annot.setdefault(img_id, list()).append(ann["id"])

            image_annot_type_map[img_id][annot_cat].update(annot_types)

        return MetricResult(
            dataset_level={"dtype": OutputsTypes.KEY_ARRAY, "subtype": "int", "value": imgs_ids_repeated_annot},
        )


__all__ = ["DatasetCompletness", "AnnotationCorrectness", "AnnotationUniqueness"]



"""Module for DOM sharpness.

Adapted from:
pydom, Sharpness Estimation for Document and Scene Images.
    https://github.com/umang-singhal/pydom


Original paper:
Kumar et al., Sharpness estimation for document and scene images (2012).
    https://ieeexplore.ieee.org/document/6460868
"""

import cv2
import numpy as np
from loguru import logger

from pymdma.common.definitions import Metric
from pymdma.common.output import DistributionResult, MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType


def _dom(median_blurred: np.ndarray):
    """Find DOM at each pixel.

    Parameters
    ----------
    median_blurred : np.ndarray
        Median filtered image.

    Returns
    -------
    domx : np.ndarray
        Diff of diff on x axis.
    domy : np.ndarray
        Diff of diff on y axis.
    """
    median_shift_up = np.pad(median_blurred, ((0, 2), (0, 0)), "constant")[2:, :]
    median_shift_down = np.pad(median_blurred, ((2, 0), (0, 0)), "constant")[:-2, :]
    domx = np.abs(median_shift_up - 2 * median_blurred + median_shift_down)

    median_shift_left = np.pad(median_blurred, ((0, 0), (0, 2)), "constant")[:, 2:]
    median_shift_right = np.pad(median_blurred, ((0, 0), (2, 0)), "constant")[:, :-2]
    domy = np.abs(median_shift_left - 2 * median_blurred + median_shift_right)
    return domx, domy


def _sharpness_matrix(
    median_blurred: np.ndarray,
    edgex: np.ndarray,
    edgey: np.ndarray,
    width: int = 2,
):
    """Find sharpness value at each pixel.

    Parameters
    ----------
    median_blurred : np.ndarray
        Median filtered grayscale image.
    edgex : np.ndarray
        Edge pixels in x-axis.
    edgey : np.ndarray
        Edge pixels in y-axis.
    width : int, optional
        Edge width, by default 2.
    debug : bool, optional
        To show intermediate results, by default False.

    Returns
    -------
    Sx : np.ndarray
        Sharpness value matrix computed in x-axis.
    Sy : np.ndarray
        Sharpness value matrix computed in y-axis.
    """
    # Compute dod measure on both axis
    domx, domy = _dom(median_blurred)

    # Contrast on x and y axis
    Cx = np.abs(median_blurred - np.pad(median_blurred, ((1, 0), (0, 0)), "constant")[:-1, :])
    Cy = np.abs(median_blurred - np.pad(median_blurred, ((0, 0), (1, 0)), "constant")[:, :-1])

    # Filter out sharpness at pixels other than edges
    Cx = np.multiply(Cx, edgex)
    Cy = np.multiply(Cy, edgey)

    # initialize sharpness matriz with 0's
    Sx = np.zeros(domx.shape)
    Sy = np.zeros(domy.shape)

    # Compute Sx
    for i in range(width, domx.shape[0] - width):
        num = np.abs(domx[i - width : i + width, :]).sum(axis=0)
        dn = Cx[i - width : i + width, :].sum(axis=0)
        Sx[i] = [(num[k] / dn[k] if dn[k] > 1e-3 else 0) for k in range(Sx.shape[1])]

    # Compute Sy
    for j in range(width, domy.shape[1] - width):
        num = np.abs(domy[:, j - width : j + width]).sum(axis=1)
        dn = Cy[:, j - width : j + width].sum(axis=1)
        Sy[:, j] = [(num[k] / dn[k] if dn[k] > 1e-3 else 0) for k in range(Sy.shape[0])]
    return Sx, Sy


def _sharpness_measure(
    median_blurred: np.ndarray,
    edgex: np.ndarray,
    edgey: np.ndarray,
    width: int,
    sharpness_threshold: float,
    epsilon: float = 1e-8,
):
    """Final Sharpness Value.

    Parameters
    ----------
    median_blurred : np.ndarray
        Median filtered grayscale image.
    width : int
        Edge width.
    sharpness_threshold : float
        Thresold to consider if a pixel is sharp.
    epsilon : float, optional
        Small value to defer div by zero, by default 1e-8.

    Returns
    -------
    S : float
        Sharpness measure(0<S<sqrt(2)).
    """
    Sx, Sy = _sharpness_matrix(median_blurred, edgex, edgey, width=width)
    Sx = np.multiply(Sx, edgex)
    Sy = np.multiply(Sy, edgey)

    Rx = np.sum(Sx >= sharpness_threshold) / (np.sum(edgex) + epsilon)
    Ry = np.sum(Sy >= sharpness_threshold) / (np.sum(edgey) + epsilon)
    return np.sqrt(Rx**2 + Ry**2)


def _smoothen_image(image: np.ndarray, transpose: bool = False, epsilon: float = 1e-8):
    """Smmoth image with ([0.5, 0, -0.5]) 1D filter.

    Parameters
    ----------
    image : np.ndarray
        Grayscale image.
    transpose : bool, optional
        To apply filter on vertical axis, by default False.
    epsilon : float, optional
        Small value to defer div by zero, by default 1e-8.

    Returns
    -------
    image_smoothed : np.ndarray
        Smoothened image.
    """
    fil = np.array([0.5, 0, -0.5])  # Smoothing Filter

    # change image axis for column convolution
    if transpose:
        image = image.T

    # Convolve grayscale image with smoothing filter
    image_smoothed = np.array([np.convolve(image[i], fil, mode="same") for i in range(image.shape[0])])

    # change image axis after column convolution
    if transpose:
        image_smoothed = image_smoothed.T

    # Normalize smoothened grayscale image
    return np.abs(image_smoothed) / (np.max(image_smoothed) + epsilon)


def _get_sharpness(
    img,
    width=2,
    sharpness_threshold=2,
    edge_threshold=0.0001,
    blur: bool = False,
    blur_size: tuple = (5, 5),
):
    """Image Sharpness Assessment.

    Parameters
    ----------
    img : str or np.ndarray
        Image source or image matrix.
    width : int, optional
        Text edge width, by default 2.
    sharpness_threshold : float, optional
        Thresold to consider if a pixel is sharp, by default 2.
    edge_threshold : float, optional
        Thresold to consider if a pixel is an edge pixel, by default 0.0001.
    debug : bool, optional
        To show intermediate results, by default False.

    Returns
    -------
    sharpness : float
        Image sharpness measure(0<S<sqrt(2)).
    """
    if len(img.shape) == 3:
        image = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    elif len(img.shape) == 2:
        image = img
    else:
        raise ValueError("Image is not in correct shape")

    # Add Gaussian Blur
    if blur:
        image = cv2.GaussianBlur(image, blur_size)

    # Perform median blur for removing Noise
    median_blurred = cv2.medianBlur(image, 3, cv2.CV_64F).astype("double") / 255.0

    # Smoothen image in x and y axis
    smoothx = _smoothen_image(image, transpose=True)
    smoothy = _smoothen_image(image)
    edgex = smoothx > edge_threshold
    edgey = smoothy > edge_threshold

    return _sharpness_measure(median_blurred, edgex, edgey, width=width, sharpness_threshold=sharpness_threshold)


class DOM(Metric):
    """Computes DOM sharpness score for an image. It is effective in detecting
    motion-blur, de-focused images or inherent properties of imaging system.

    **Objective**: Sharpness

    Parameters
    ----------
    width : int, optional, default=2
        Width of the edge filter.
    sharpness_threshold : int, optional, default=2
        Threshold for considering if a pixel is sharp or not.
    edge_threshold : float, optional, default=0.0001
        Threshold for edge.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    Kumar et al., Sharpness estimation for document and scene images (2012).
    https://ieeexplore.ieee.org/document/6460868

    Code was adapted from:
    pydom, Sharpness Estimation for Document and Scene Images.
    https://github.com/umang-singhal/pydom

    Examples
    --------
    >>> dom = DOM()
    >>> imgs = np.random.rand(20, 100, 100, 3) # (N, H, W, C)
    >>> result: MetricResult = dom.compute(imgs)
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.INSTANCE
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        width: int = 2,
        sharpness_threshold: int = 2,
        edge_threshold: float = 0.0001,
        blur: bool = False,
        blur_size: tuple = (5, 5),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.width = width
        self.sharpness_threshold = sharpness_threshold
        self.edge_threshold = edge_threshold
        self.blur = blur
        self.blur_size = blur_size
        if blur:
            logger.warning("Applying Gaussian Blur to the images may lead to non-deterministic results.")

    def compute(
        self,
        imgs: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes DOM score for an image.

        Parameters
        ----------
        imgs : {(N, H, W, C) ndarray, (N, H, W) ndarray}
            List of arrays representing RGB or grayscale image of shape (H, W, C) or (H, W), respectively.

        Returns
        -------
        result: MetricResult
            DOM score for each image.
        """
        scores = [
            _get_sharpness(
                img,
                self.width,
                self.sharpness_threshold,
                self.edge_threshold,
                self.blur,
                self.blur_size,
            )
            for img in imgs
        ]

        return DistributionResult(
            instance_level={"dtype": OutputsTypes.ARRAY, "subtype": "float", "value": scores},
        )


__all__ = ["DOM"]



from typing import List, Optional, Union

import numpy as np
import ot
from scipy.signal import coherence, welch

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType


class SpectralCoherence(Metric):
    """Compute the mean coherence between real and synthetic signals. Returns
    the average (median-based) coherence across signals.

    **Objective**: Similarity

    Parameters
    ----------
    fs: int, optional, default=2048
        The sampling frequency of the signal.
    nperseg: int, optional, default=None
        Length of each segment used to compute the power spectral density.
    valid_freq_range: tuple, optional, default=(-np.inf, np.inf)
        The range of valid frequencies for the signal.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    F. P. Carrle, Y. Hollenbenders, and A. Reichenbach, Generation of synthetic EEG data for training algorithms supporting the diagnosis of major depressive disorder (2023).
    https://www.frontiersin.org/journals/neuroscience/articles/10.3389/fnins.2023.1219133/full.

    J. Vetter, J. H. Macke, and R. Gao, Generating realistic neurophysiological time series with denoising diffusion probabilistic models (2024)
    https://pmc.ncbi.nlm.nih.gov/articles/PMC11573898/

    A. Zancanaro, I. Zoppis, S. Manzoni, and G. Cisotto, vEEGNet: A New Deep Learning Model to Classify and Generate EEG (2023)
    https://www.scitepress.org/DigitalLibrary/Link.aspx?doi=10.5220/0011990800003476.

    Examples
    --------
    >>> coeherence = SpectralCoherence()
    >>> real_data = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> fake_data = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = coherence.compute(real_data, fake_data)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        fs: int = 2048,
        nperseg: Optional[int] = None,
        valid_freq_range: tuple = (-np.inf, np.inf),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.fs = fs
        self.valid_freq_range = valid_freq_range
        self.nperseg = nperseg

    def _preprocess(self, data: Union[np.ndarray, List[np.ndarray]]):
        """Preprocess data for computation.

        Parameters
        ----------
        data : Union[np.ndarray, List[np.ndarray]]
            Input data, either a 1D, 2D, or 3D array.

        Returns
        -------
        data : np.ndarray
            Preprocessed data, a 2D array.
        """
        data = np.array(data)
        # 1D signals need to be converted to 2D
        if data.ndim == 1:
            data = np.expand_dims(data, axis=1)
        elif data.ndim > 2:
            # Convert to 2D
            data = data.reshape(data.shape[0], -1)
        return data

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Compute the metric.

        Parameters
        ----------
        real_features : array-like of shape (N, L) or (N, L, C)
            Array with the original samples.
        fake_features : array-like of shape (N, L) or (N, L, C)
            Array with the fake samples.
        **kwargs : dict, optional
            Additional keyword arguments for compatibility (unused).

        Returns
        -------
        MetricResult
            Dataset-level results for the Spectral Coherence.
        """

        coherence_values = []

        real_data = self._preprocess(real_data)
        syn_data = self._preprocess(syn_data)

        # Pair up signals one-to-one
        for real_sig, synth_sig in zip(real_data, syn_data):
            win = min(len(real_sig) // 8, self.fs // 2) if self.nperseg is None else self.nperseg
            f, Cxy = coherence(real_sig, synth_sig, fs=self.fs, nperseg=win, window="boxcar")
            valid_freqs = (f >= self.valid_freq_range[0]) & (f <= self.valid_freq_range[1])
            coherence_values.append(np.median(Cxy[valid_freqs]))

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.NUMERIC,
                "subtype": "float",
                "value": np.mean(coherence_values),
            },
        )


class SpectralWassersteinDistance(Metric):
    """Compute a Wasserstein distance in the frequency domain based on Power
    Spectral Density (PSD) differences.

    **Objective**: Similarity

    Parameters
    ----------
    fs: int, optional, default=2048
        The sampling frequency of the signal.
    nperseg: int, optional, default=None
        Length of each segment used to compute the power spectral density.
    **kwargs : dict, optional
        Additional keyword arguments for compatibility.

    References
    ----------
    F. P. Carrle, Y. Hollenbenders, and A. Reichenbach, Generation of synthetic EEG data for training algorithms supporting the diagnosis of major depressive disorder (2023).
    https://www.frontiersin.org/journals/neuroscience/articles/10.3389/fnins.2023.1219133/full.

    Examples
    --------
    >>> spectral_wd = SpectralWassersteinDistance()
    >>> real_data = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> fake_data = np.random.rand(64, 1000, 12) # (N, L, C)
    >>> result: MetricResult = spectral_wd.compute(real_data, fake_data)
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = 0.0
    max_value: float = np.inf

    def __init__(
        self,
        fs: int = 2048,
        nperseg: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.fs = fs
        self.nperseg = nperseg

    def _compute_power_spectral_density(self, data) -> np.ndarray:
        """Compute the power spectral density of the given data.

        Parameters
        ----------
        data: array-like of shape (N,) or (N, L) or (N, L, C)
            The input data.

        Returns
        -------
        psd: array-like of shape (N, L) or (N, L, C)
            The power spectral density of the input data.
        """
        if isinstance(data, np.ndarray) and data.ndim == 1:
            data = [data]

        psd = []
        for sig in data:
            # Use a segment length (nperseg) to compute PSD:
            win = min(len(sig), 4 * self.fs) if self.nperseg is None else self.nperseg
            psd.append(welch(sig, self.fs, nperseg=win, window="boxcar")[1])
        return np.array(psd)

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Compute the metric.

        Parameters
        ----------
        real_features : array-like of shape (N, L) or (N, L, C)
            Array with the original samples.
        fake_features : array-like of shape (N, L) or (N, L, C)
            Array with the fake samples.
        **kwargs : dict, optional
            Additional keyword arguments for compatibility (unused).

        Returns
        -------
        MetricResult
            Dataset-level results for the Spectral Coherence.
        """
        real_psd = self._compute_power_spectral_density(real_data)
        syn_psd = self._compute_power_spectral_density(syn_data)

        # Align number of frequency bins
        min_len = min(real_psd.shape[1], syn_psd.shape[1])
        real_psd = real_psd[:, :min_len]
        syn_psd = syn_psd[:, :min_len]

        real_psd = real_psd.reshape(-1, min_len)
        syn_psd = syn_psd.reshape(-1, min_len)

        # Cost matrix in PSD space (Euclidean)
        cost_matrix_psd = ot.dist(real_psd, syn_psd, metric="euclidean")
        # Uniform distributions
        a = np.ones(real_psd.shape[0]) / real_psd.shape[0]
        b = np.ones(syn_psd.shape[0]) / syn_psd.shape[0]

        # Sinkhorn regularized Wasserstein distance
        wd_psd = ot.sinkhorn2(a, b, cost_matrix_psd, reg=0.01)

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.NUMERIC,
                "subtype": "float",
                "value": wd_psd,
            },
        )


__all__ = ["SpectralCoherence", "SpectralWassersteinDistance"]



from pathlib import Path
from typing import Optional, Union

import numpy as np
from sklearn.cluster import MiniBatchKMeans
from torch import Tensor, tensor


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)


def set_seed(seed: int = 42):
    """Set random seed for numpy.

    https://towardsdatascience.com/stop-using-numpy-random-seed-581a9972805f
    """
    rng = np.random.default_rng(seed)
    return rng


def check_input_shape(dataset, ndim=2):
    if not isinstance(dataset, np.ndarray) or dataset.ndim != ndim:
        raise ValueError(
            "Input variable is not a NumPy array with the required dimensions",
        )


# ======= General Functions =======
def min_max_scaling(array: np.ndarray) -> np.ndarray:
    return (array - array.min()) / (array.max() - array.min())


def features_splitting(feat, seed: Optional[int] = None):
    """Split the input array into two subsets with equal lengths. If the number
    of elements in the input array is odd, the function discards one element to
    ensure both subsets have the same length.

    Parameters
    ----------
    feat : np.ndarray (n_samples, n_features)
        Input array to be split. The first dimension should be the number of elements.

    seed : int, optional
        Seed for the random number generator, ensuring reproducibility of the split.
        If None (default), the random number generator uses the current random state.

    Returns
    -------
    tuple
        A tuple of two np.ndarray objects, each representing a subset of the input array.
    """
    rng = set_seed(seed)
    # Randomly sample two subsets
    subset_indices1 = rng.choice(feat.shape[0], size=len(feat) // 2, replace=False)
    remaining_indices = np.setdiff1d(np.arange(len(feat)), subset_indices1)
    subset_indices2 = rng.choice(remaining_indices, size=len(feat) // 2, replace=False)

    # Use the sampled indices to extract subsets from the original array
    return feat[subset_indices1], feat[subset_indices2]


def cluster_into_bins(eval_data, ref_data, num_clusters):
    """Clusters the union of the data points and returns the cluster
    distribution.

    Clusters the union of eval_data and ref_data into num_clusters using minibatch
    k-means. Then, for each cluster, it computes the number of points from

    Parameters
    ----------
    eval_data : np.ndarray
        Data points from the distribution to be evaluated.
    ref_data : np.ndarray
        Data points from the reference distribution.
    num_clusters : int
        Number of cluster centers to fit.

    Returns
    -------
    eval_bins : np.ndarray
        Array where the i-th entry represents the number of points from eval_data assigned to the i-th cluster.
    ref_bins : np.ndarray
        Array where the i-th entry represents the number of points from ref_data assigned to the i-th cluster.

    References
    ---------
    Sajjadi, Mehdi SM, et al. Assessing generative models via precision and recall (2018).
    https://proceedings.neurips.cc/paper_files/paper/2018/file/f7696a9b362ac5a51c3dc8f098b73923-Paper.pdf

    Code taken from:
    https://github.com/vanderschaarlab/evaluating-generative-models/blob/main/metrics/prd_score.py
    """

    cluster_data = np.vstack([eval_data, ref_data])
    kmeans = MiniBatchKMeans(n_clusters=num_clusters, n_init=10, random_state=42)
    labels = kmeans.fit(cluster_data).labels_

    eval_labels = labels[: len(eval_data)]
    ref_labels = labels[len(eval_data) :]

    eval_bins = np.histogram(
        eval_labels,
        bins=num_clusters,
        range=[0, num_clusters],
        density=True,
    )[0]
    ref_bins = np.histogram(
        ref_labels,
        bins=num_clusters,
        range=[0, num_clusters],
        density=True,
    )[0]
    return eval_bins, ref_bins


def to_tensor(data: Union[Tensor, np.ndarray]) -> Tensor:
    return data if isinstance(data, Tensor) else tensor(data)


# EOF



from .imagenet import DinoExtractor, InceptionFID, VGGExtractor, ViTExtractor


class ExtractorFactory:
    default = "vit_b_32"

    @staticmethod
    def model_from_name(
        name: str,
    ):
        """Initializes the feature extractor with the given parameters.

        Args:
            name (str): identifier of the extractor to be used.
            device (str): model device. Defaults to "cpu".
        """
        # name = self.default if name == "default" else name

        if name == "inception_fid":
            extractor = InceptionFID()
        elif "vgg" in name:
            extractor = VGGExtractor(name)
        elif "dino" in name:
            extractor = DinoExtractor(name)
        elif "vit" in name:
            extractor = ViTExtractor(name)
        else:
            raise ValueError(f"Model {name} not available.")

        return extractor.eval()



from typing import List, Optional

import numpy as np

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ...utils_inp import compute_k_anonymity


class KAnonymityScore(Metric):
    """Calculates the k for k-anonymity. A higher k value indicates that each
    record is less unique, meaning it is more difficult to re-identify
    individuals within the dataset.

    **Objective**: Privacy

    Parameters
    ----------
    column_names : list
        List of the names of the columns (features) in the dataset.
    qi_names : list, optional, default=None
        List of the quasi-identifier column names.
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Díaz and García, A python library to check the level of anonymity of a dataset. (2022).
    http://dx.doi.org/10.1038/s41597-022-01894-2

    Returns
    -------
    MetricResult
        A MetricResult object containing the k-anonymity score.

    Examples
    --------
    >>> # Example 1: Evaluating k-anonymity on a dataset with sufficient quasi-identifiers
    >>> import numpy as np
    >>> data = np.array([
    ...     ['Alice', 'Smith', 'NY'],
    ...     ['Alice', 'Smith', 'NY'],
    ...     ['Bob', 'Jones', 'CA'],
    ...     ['Bob', 'Jones', 'CA']
    ... ])
    >>> k_anonymity = KAnonymityScore(
    ...     column_names=['first_name', 'last_name', 'state'],
    ...     qi_names=['first_name', 'last_name']
    ... )
    >>> result = k_anonymity.compute(data)
    >>> dataset_level, _ = result.value # Output: k-anonymity score

    >>> # Example 2: Evaluating k-anonymity on a dataset with low uniqueness
    >>> data = np.array([
    ...     ['Alice', 'Smith', 'NY'],
    ...     ['Alice', 'Smith', 'CA'],
    ...     ['Bob', 'Jones', 'NY']
    ... ])
    >>> k_anonymity = KAnonymityScore(
    ...     column_names=['first_name', 'last_name', 'state'],
    ...     qi_names=['first_name', 'last_name']
    ... )
    >>> result: MetricResult = k_anonymity.compute(data)
    >>> dataset_level, _ = result.value # Output: k-anonymity score
    """

    reference_type = ReferenceType.NONE
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.PRIVACY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 100.0

    def __init__(self, column_names: Optional[List[str]] = None, qi_names: Optional[List[str]] = None, **kwargs):
        super().__init__(**kwargs)
        self.column_names = column_names
        self.qi_names = qi_names

    def compute(self, data: np.ndarray, **kwargs) -> MetricResult:
        """Computes the k-anonymity score for the dataset based on the
        specified quasi-identifiers.

        Parameters
        ----------
        data : np.ndarray
            The input dataset for which k-anonymity will be computed. The dataset
            should be in array-like format with the shape (samples, features).
        **kwargs : dict
            Additional keyword arguments for controlling the computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the k-anonymity score.
        """

        # columns
        cols = (
            self.column_names
            if isinstance(self.column_names, list)
            else [f"att_{idx}" for idx in range(data.shape[-1])]
        )

        # qi names
        qi_names = self.qi_names if isinstance(self.qi_names, list) else cols[-1:]

        # k anonimity
        k_anom = compute_k_anonymity(
            data=data,
            column_names=cols,
            qi_names=qi_names,
        )

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.NUMERIC,
                "subtype": "float",
                "value": k_anom,
            },
        )


__all__ = ["KAnonymityScore"]



import numpy as np


class GenericEmbedder:
    def __init__(self, **kwargs) -> None:
        # has inverse transform
        self.has_inv = False

    def fit(self, X: np.ndarray, **kwargs):
        pass

    def transform(self, X: np.ndarray, **kwargs):
        return X

    def fit_transform(self, X: np.ndarray, **kwargs):
        # fit
        self.fit(X)

        # transform
        X_new = self.transform(X)

        return X_new

    def inverse_transform(self, X: np.ndarray, **kwargs):
        return X


class GenericNorm:
    def __init__(self, **kwargs):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X

    def fit_transform(self, X, y=None):
        return X

    def inverse_transform(self, X):
        return X



from pathlib import Path
from typing import Callable, List, Optional, Tuple

import numpy as np
import torch
from loguru import logger
from PIL import Image
from torch.utils.data import DataLoader
from torchvision.transforms import transforms
from tqdm import tqdm

from pymdma.common.definitions import EmbedderInterface

# support for large datasets iwht small batch sizes
torch.multiprocessing.set_sharing_strategy("file_system")


class StandardTransform:
    def __init__(
        self,
        image_size: Tuple[int],
        interpolation: Image.Resampling = Image.Resampling.BILINEAR,
        preprocess_transform: Optional[Callable] = None,
    ) -> None:
        assert isinstance(image_size, tuple), "Image size must be a tuple."
        self.img_size = image_size
        self.interp = interpolation
        self.preprocess_transform = preprocess_transform

        self._to_tensor = transforms.PILToTensor()

    def __call__(self, image: Image.Image) -> torch.Tensor:
        image = self.preprocess_transform(image) if self.preprocess_transform is not None else image
        image = image.resize(self.img_size, self.interp)
        # bring image to the range [0, 1] and normalize to [-1, 1]
        image = self._to_tensor(image).float() / 255.0
        image = image * 2.0 - 1.0
        return image


class BaseExtractor(torch.nn.Module, EmbedderInterface):
    extractor: torch.nn.Module

    def __init__(
        self,
        input_size: Tuple[int],
        interpolation: Image.Resampling.BILINEAR,
    ):
        super().__init__()
        self.input_size = input_size
        self.interpolation = interpolation

    @torch.no_grad()
    def extract_features_from_files(
        self,
        files: List[Path],
        batch_size: int = 50,
        device: str = "cpu",
        preprocess_transform: Optional[Callable] = None,
    ) -> np.ndarray:
        """Extract features from a list of image files. Converts images to
        tensors and normalizes them to the [-1, 1] range.

        Parameters
        ----------
        files : List[Path]
            list of paths to image files
        batch_size : int, optional
            batch size for feature extraction, by default 50

        Returns
        -------
        np.ndarray
            array of features with shape (n_samples, n_features)
        """
        if batch_size > len(files):
            # print("Warning: batch size is bigger than the data size. " "Setting batch size to data size")
            batch_size = len(files)

        assert len(files) > 0, "No files to extract features from."

        n_batches = len(files) // batch_size
        batch_sizes = [batch_size for _ in range(n_batches)]
        if len(files) % batch_size != 0:
            batch_sizes.append(len(files) % batch_size)

        transform = StandardTransform(self.input_size, self.interpolation, preprocess_transform)
        self.extractor = self.extractor.to(device, dtype=torch.float32)

        act_array = []
        start, end = 0, 0
        for bsize in batch_sizes:
            end = start + bsize
            images = [transform(Image.open(f).convert("RGB")) for f in files[start:end]]
            batch = torch.stack(images).to(device)
            batch = self(batch).detach().cpu().numpy()
            act_array.append(batch)
            start += bsize
        return np.concatenate(act_array, axis=0)

    @torch.no_grad()
    def extract_features_from_dataloader(
        self,
        dataloader: DataLoader,
        normalize: bool = False,
        device: str = "cpu",
    ):
        """Extract features from a DataLoader.

        Parameters
        ----------
        dataloader : DataLoader
            PyTorch DataLoader
        normalize : bool, optional
            Wether to normalize the images to 0.5 mean and 0.5 std across channels.
            Assumes that the images are in the [0, 1] range, by default False
        device : str, optional
            device to use, by default "cpu"

        Notes
        -----
        Has the following assumptions:
         - Dataloader outputs a tuple with the first element being the image batch tensor.
         - Image tensors are of shape (batch_size, channels, height, width)
         - Tensors are in the [0, 1] range in which case you should set `normalize` to True.
         - Tensors have been normalized to the [-1, 1] range (0.5 mean and 0.5 std across channels) or to another range (e.g. ImageNet normalization).

        It is recommended to disable the `shuffle` option in the DataLoader for consistency.

        Depending on the model, you might need to resize the images to the model's input size.

        Returns
        -------
        np.ndarray
            array of features with shape (n_samples, n_features)
        """
        self.extractor = self.extractor.to(device, dtype=torch.float32)

        # validation dry run
        sample_batch = next(iter(dataloader))
        assert isinstance(
            sample_batch[0], torch.Tensor
        ), f"First element of the tuple must be a torch.Tensor. Got {type(sample_batch[0])}."

        if sample_batch[0].shape[2:] != self.input_size:
            logger.warning(
                f"Model default input size {self.input_size} does not match the size of the images in the dataloader {sample_batch[0][2:]}. Might lead to execution errors."
            )

        act_array = []
        for batch in tqdm(dataloader, total=len(dataloader)):
            images = batch[0]
            images = images.to(device, dtype=torch.float32)
            if normalize:
                images = images * 2.0 - 1.0
            act_array.append(self(images).detach().cpu().numpy())
        return np.concatenate(act_array, axis=0)

    @torch.no_grad()
    def _extract_features_dataloader(
        self,
        dataloader: DataLoader,
        device: str = "cpu",
        preprocess_transform: Optional[Callable] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Internal method to extract features from a DataLoader.

        Parameters
        ----------
        dataloader : DataLoader
            image dataloader
        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: extracted features, labels, and image ids
        """
        logger.info("Extracting image features.")
        act_array = []
        labels_array = []
        ids_array = []

        self.extractor = self.extractor.to(device, dtype=torch.float32)
        dataloader.dataset.transform = StandardTransform(self.input_size, self.interpolation, preprocess_transform)
        for batch, labels, img_ids in tqdm(dataloader, total=len(dataloader)):
            batch = batch.to(device, dtype=torch.float32)
            batch = self(batch).detach().cpu().numpy()
            act_array.append(batch)
            labels_array.append(labels)
            ids_array.append(img_ids)

        return (
            np.concatenate(act_array, axis=0),
            np.concatenate(labels_array, axis=0),
            np.concatenate(ids_array, axis=0),
        )

    def forward(self, x):
        return self.extractor(x)



from typing import Dict, List, Literal, Optional

import numpy as np
import pandas as pd

from pymdma.common.definitions import Metric
from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, OutputsTypes, ReferenceType

from ....data.utils import is_categorical
from ...utils_syn import _get_js_divergence, _get_kl_divergence, _get_ks_similarity, _get_nn_pdf, _get_tv_similarity


class StatisticalSimScore(Metric):
    """Computes a dataset-level statistical similarity score between real and
    synthetic data.

    This metric assesses how closely the statistical properties of the synthetic dataset
    resemble those of the real dataset, providing a fidelity measure for synthetic data generation.

    **Objective**: Fidelity

    Parameters
    ----------
    col_map : dict, optional, default=None
        A mapping of column names to their types and properties. This is used to determine
        how to compute similarity for each column.
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Yang et al., Structured evaluation of synthetic tabular data (2024).
    https://arxiv.org/abs/2403.10424

    Returns
    -------
    MetricResult
        A MetricResult object containing the similarity scores and their statistics.

    Examples
    --------
    >>> # Example 1: Evaluating statistical similarity for a dataset with discrete and continuous variables
    >>> import numpy as np
    >>> real_data = np.array([
    ...     [1, 2.5],
    ...     [1, 3.0],
    ...     [2, 3.5],
    ...     [2, 4.0]
    ... ])
    >>> syn_data = np.array([
    ...     [1, 2.6],
    ...     [1, 3.1],
    ...     [2, 3.4],
    ...     [2, 4.2]
    ... ])
    >>> col_map = {
    ...     "column1": {"type": {"tag": "discrete"}},
    ...     "column2": {"type": {"tag": "continuous"}},
    ... }
    >>> sim_score = StatisticalSimScore(col_map=col_map)
    >>> result: MetricResult = sim_score.compute(real_data, syn_data)
    >>> dataset_level, _ = result.value # Output: similarity scores for each column

    >>> # Example 2: Evaluating similarity with mismatched column types
    >>> real_data = np.array([
    ...     [1, 2],
    ...     [2, 3],
    ...     [3, 4]
    ... ])
    >>> syn_data = np.array([
    ...     [1, 2],
    ...     [2, 3],
    ...     [3, 5]
    ... ])
    >>> col_map = {
    ...     "column1": {"type": {"tag": "discrete"}},
    ...     "column2": {"type": {"tag": "discrete"}},
    ... }
    >>> sim_score = StatisticalSimScore(col_map=col_map)
    >>> result: MetricResult = sim_score.compute(real_data, syn_data)
    >>> dataset_level, _ = result.value  # Output: similarity scores for each column
    >>> dataset_stats, _ = result.stats  # Output: mean and std of similarity scores
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        col_map: Optional[Dict[str, Dict[str, str]]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.col_map = col_map

    def _stat_sim_1d(
        self,
        real_col: np.ndarray,
        syn_col: np.ndarray,
        kind: str = "discrete",
        **kwargs,
    ):
        """Computes an 1D statistical similarity.

        Parameters
        ----------
        real_col : np.ndarray
            The real data for the specific attribute.
        syn_col : np.ndarray
            The synthetic data for the specific attribute.
        kind : str
            The type of data ('discrete' or 'continuous').
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        float
            The computed similarity score for the attribute.
        """
        # variable type assignment
        kind_ = kind.lower() if kind.lower() in ["discrete", "continuous"] else "continuous"

        # mapper
        kind_mapper = {
            "discrete": _get_tv_similarity,
            "continuous": _get_ks_similarity,
        }

        # score
        sim_score = kind_mapper.get(kind_)(
            real_col,
            syn_col,
        )
        return sim_score

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Computes the statistical similarity score between real and synthetic
        datasets.

        Parameters
        ----------
        real_data : np.ndarray
            The real dataset for comparison.
        syn_data : np.ndarray
            The synthetic dataset to evaluate.
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the similarity scores and their statistics.
        """

        # checkpoint
        assert real_data.shape[1] == syn_data.shape[1], "Mismatched columns. Please fix before computing metrics."

        # column map
        col_map_exists = isinstance(self.col_map, dict)
        cols = self.col_map.keys() if col_map_exists else [f"att_{idx}" for idx in range(real_data.shape[1])]

        # similarity map
        sim_score = {}

        # column similarity
        for idx, col in enumerate(cols):
            # continuous OR discrete
            if col_map_exists:
                # dtype
                vtype = self.col_map.get(col).get("type").get("tag")
            else:
                # dtype
                vtype = "discrete" if is_categorical(real_data[:, idx]) else "continuous"

            # compute similarity
            sim_ = self._stat_sim_1d(
                real_data[:, idx],
                syn_data[:, idx],
                kind=vtype,
            )

            # assign
            sim_score[col] = sim_

        # global scores
        global_d = {
            "mean": np.mean(list(sim_score.values())).round(2),
            "std": np.std(list(sim_score.values())).round(2),
        }

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_VAL,
                "subtype": "float",
                "value": sim_score,
                "stats": global_d,
            },
        )


class StatisticalDivergenceScore(Metric):
    """Computes a statistical divergence score for each column, specifically
    the Jensen-Shannon (JS) and Kullback-Leibler (KL) divergence scores.

    **Objective**: Fidelity

    Parameters
    ----------
    column_names : list of str, optional, default=None
        List of the names of the columns (features) in the dataset.
    score : str, optional, default='kl'
        Specifies the divergence score to compute ('js' for Jensen-Shannon, 'kl' for Kullback-Leibler, 'all' for both).
        By default, it is set to 'kl'.
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Fonseca and Bacao,  Tabular and latent space synthetic data generation: a literature review (2023).
    https://doi.org/10.1186/s40537-023-00792-7

    Returns
    -------
    MetricResult
        A MetricResult object containing the divergence scores and their statistics.

    Examples
    --------
    >>> # Example 1: Evaluating statistical divergence for a dataset
    >>> import numpy as np
    >>> real_data = np.array([
    ...     [1, 2, 3],
    ...     [2, 3, 4],
    ...     [3, 4, 5]
    ... ])
    >>> syn_data = np.array([
    ...     [1, 2, 2],
    ...     [2, 2, 3],
    ...     [3, 3, 4]
    ... ])
    >>> col_map = {
    ...     "column1": {"type": {"tag": "continuous"}},
    ...     "column2": {"type": {"tag": "continuous"}},
    ...     "column3": {"type": {"tag": "continuous"}},
    ... }
    >>> divergence_score = StatisticalDivergenceScore(col_map=col_map, score='kl')
    >>> result: MetricResult = divergence_score.compute(real_data, syn_data)
    >>> dataset_level, _ = result.value  # Output: divergence scores for each column

    >>> # Example 2: Using JS divergence instead of KL
    >>> divergence_score_js = StatisticalDivergenceScore(col_map=col_map, score='js')
    >>> result_js: MetricResult = divergence_score_js.compute(real_data, syn_data)
    >>> dataset_level_js, _ = result_js.value  # Output: JS divergence scores for each column
    >>> dataset_stats_js, _ = result_js.stats  # Output: mean and std of JS divergence scores
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = False
    min_value: float = -np.inf
    max_value: float = np.inf

    def __init__(
        self,
        column_names: Optional[List[str]] = None,
        score: Literal["js", "kl", "all"] = "kl",
        **kwargs,
    ):
        """Initializes the StatisticalDivergenceScore metric to evaluate the
        divergence between real and synthetic datasets based on defined column
        characteristics.

        Parameters
        ----------
        column_names : list of str, optional, default=None
            List of the names of the columns (features) in the dataset.
        score : str, optional, default='kl'
            Specifies the divergence score to compute ('js' for Jensen-Shannon,
            'kl' for Kullback-Leibler). Default is 'kl'.
        **kwargs : dict
            Additional keyword arguments passed to the parent class.
        """

        super().__init__(**kwargs)

        # column names
        self.column_names = column_names

        # score type
        self.score = score

    def _diverg_score_1d(
        self,
        real_col: np.ndarray,
        syn_col: np.ndarray,
        score: Literal["js", "kl", "all"] = "all",
        **kwargs,
    ):
        """Computes a column-level statistical divergence.

        Parameters
        ----------
        real_col : np.ndarray
            The real data for the specific attribute.
        syn_col : np.ndarray
            The synthetic data for the specific attribute.
        score : str
            The type of divergence to compute ('js' or 'kl').
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        dict
            A dictionary containing the computed divergence scores.
        """

        # score map
        score_map = {"js": _get_js_divergence, "kl": _get_kl_divergence}

        # score tags
        score_tag = ["js", "kl"] if score.lower() == "all" else [score]

        # get probability distributions
        real_pdf, syn_pdf, bins = _get_nn_pdf(real_col, syn_col)

        # compute divergence scores
        div_score = {tag: score_map.get(tag)(real_pdf, syn_pdf) for tag in score_tag if tag in score_map.keys()}

        return div_score

    def compute(self, real_data: np.ndarray, syn_data: np.ndarray, **kwargs) -> MetricResult:
        """Computes the statistical divergence score between real and synthetic
        datasets.

        Parameters
        ----------
        real_data : np.ndarray
            The real dataset for comparison.
        syn_data : np.ndarray
            The synthetic dataset to evaluate.
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the divergence scores and their statistics.
        """

        # checkpoint
        assert real_data.shape[1] == syn_data.shape[1], "Mismatched columns. Please fix before computing metrics."

        # columns
        cols = (
            self.column_names
            if isinstance(self.column_names, list)
            else [f"att_{idx}" for idx in range(real_data.shape[1])]
        )

        # divergence map
        div_score = {}

        # column-wise
        for idx, col in enumerate(cols):
            # compute scores
            sim_ = self._diverg_score_1d(
                real_data[:, idx],
                syn_data[:, idx],
                score=self.score,
            )

            # assign
            div_score[col] = list(sim_.values())

        # global scores
        # auxiliary score dataframe
        aux_df = pd.DataFrame.from_dict(div_score.values())

        # aggregates
        mean_g, std_g = aux_df.mean(0).to_dict(), aux_df.std(0).to_dict()

        # global
        glob_d = {
            f"{self.score}_mean": mean_g[0],
            f"{self.score}_std": std_g[0],
        }

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.KEY_ARRAY,
                "subtype": "float",
                "value": div_score,
                "stats": glob_d,
            },
        )


class CoherenceScore(Metric):
    """Computes the coherence score between the correlation matrices of the
    target and synthetic datasets. A higher coherence score indicates better
    fidelity between the datasets in terms of their correlation structures.

    **Objective**: Fidelity

    Parameters
    ----------
    weights : np.ndarray, optional, default=None
        Weights for the correlations, allowing for weighted contributions
        to the coherence score. If None, uniform weights are applied.
    corr_type : str, optional, default='pearson'
        The type of correlation to compute ('pearson' by default).
        Other types like 'spearman' may be supported depending on the implementation.
    **kwargs : dict
        Additional keyword arguments passed to the parent class.

    References
    ----------
    Yang et al., Structured evaluation of synthetic tabular data (2024).
    https://arxiv.org/abs/2403.10424

    Returns
    -------
    MetricResult
        A MetricResult object containing the coherence score.

    Examples
    --------
    >>> # Example 1: Evaluating coherence score for a dataset
    >>> import numpy as np
    >>> real_data = np.array([
    ...     [1, 2, 3],
    ...     [2, 3, 4],
    ...     [3, 4, 5]
    ... ])
    >>> syn_data = np.array([
    ...     [1, 2, 3],
    ...     [1, 2, 3],
    ...     [3, 4, 5]
    ... ])
    >>> coherence_score = CoherenceScore(corr_type='pearson')
    >>> result: MetricResult = coherence_score.compute(real_data, syn_data)
    >>> dataset_level, _ = result.value  # Output: coherence score

    >>> # Example 2: Evaluating with custom weights
    >>> weights = np.array([0.5, 1.0, 1.5])  # Example weights
    >>> coherence_score_weighted = CoherenceScore(weights=weights, corr_type='spearman')
    >>> result_weighted: MetricResult = coherence_score_weighted.compute(real_data, syn_data)
    >>> dataset_level, _ = result_weighted.value  # Output: weighted coherence score
    """

    reference_type = ReferenceType.DATASET
    evaluation_level = EvaluationLevel.DATASET
    metric_group = MetricGroup.QUALITY

    higher_is_better: bool = True
    min_value: float = 0.0
    max_value: float = 1.0

    def __init__(
        self,
        weights: Optional[np.ndarray] = None,
        corr_type: Optional[str] = "pearson",
        **kwargs,
    ):
        """Initializes the CoherenceScore metric to evaluate the coherence
        between the correlation matrices of real and synthetic datasets.

        Parameters
        ----------
        weights : np.ndarray, optional, default=None
            Weights for the correlations, allowing for weighted contributions
            to the coherence score. If None, uniform weights are applied.
        corr_type : str, optional, default=None
            The type of correlation to compute ('pearson' by default).
        **kwargs : dict
            Additional keyword arguments passed to the parent class.
        """

        super().__init__(**kwargs)

        # correlation type
        self.corr = corr_type

        # weights array
        self.weights = weights

    def compute(
        self,
        real_data: np.ndarray,
        syn_data: np.ndarray,
        **kwargs,
    ) -> MetricResult:
        """Computes the coherence score between the correlation matrices of
        real and synthetic datasets.

        Parameters
        ----------
        real_data : np.ndarray
            The real dataset for comparison.
        syn_data : np.ndarray
            The synthetic dataset to evaluate.
        **kwargs : dict
            Additional keyword arguments for computation.

        Returns
        -------
        MetricResult
            A MetricResult object containing the coherence score.
        """

        # compute correlation matrices
        real_corr = pd.DataFrame(real_data).corr(self.corr).replace(np.nan, 1).to_numpy()
        syn_corr = pd.DataFrame(syn_data).corr(self.corr).replace(np.nan, 1).to_numpy()

        # number columns
        n_cols = len(real_corr)

        # compute similarity between real and syn matrices
        delta_corr = np.abs(
            np.nan_to_num(real_corr) - np.nan_to_num(syn_corr),
        )  # differences

        # weight matrix
        id_mask = np.abs(np.identity(n_cols) - 1)

        if self.weights is not None:
            w_mask = np.array([self.weights] * n_cols) * id_mask
        else:
            w_mask = np.ones((n_cols, n_cols)) * id_mask

        # norm
        w_mask /= sum(w_mask)

        # correlation similarity (weighted avg.)
        corr_sim = np.sum(delta_corr * w_mask) / np.sum(w_mask)
        # ((n_cols * (n_cols - 1)))

        # average correlation
        avg_corr_coh = np.mean(np.round(1 - corr_sim / 2, 3))

        return MetricResult(
            dataset_level={
                "dtype": OutputsTypes.NUMERIC,
                "subtype": "float",
                "value": avg_corr_coh,
            },
        )


__all__ = ["StatisticalSimScore", "StatisticalDivergenceScore", "CoherenceScore"]



from pathlib import Path
from typing import Callable, List, Optional

import numpy as np
from PIL import Image
from torch.utils.data import DataLoader, Dataset

# TODO support other formats?
SUPPORTED_FILES = {".png", ".jpg", ".jpeg"}


class SimpleDataset(Dataset):
    def __init__(
        self,
        file_paths: List[Path],
        transforms: Optional[Callable] = None,
        target_transforms: Optional[Callable] = None,
        train_transforms: Optional[Callable] = None,
    ) -> None:
        self.img_files = file_paths
        self.transform = transforms
        self.target_transform = target_transforms
        self.train_transform = train_transforms

    def __len__(self):
        return len(self.img_files)

    def __getitem__(self, idx):
        img_path = self.img_files[idx]
        image = Image.open(img_path).convert("RGB")
        image_id = Path(img_path).stem

        label = 0
        if self.transform:
            image = self.train_transform(image) if self.train_transform else image
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
        return image, label, image_id


def build_img_dataloader(
    file_paths: List[Path],
    transforms: Optional[Callable] = None,
    batch_size: int = 10,
    num_workers: int = 0,
    **kwargs,
) -> DataLoader:
    """Builds a PyTorch DataLoader for a list of image files.

    Parameters
    ----------
    file_paths : List[Path]
        List of image file paths.
    transforms : Optional[Callable], optional
        Transforms to be applied to the images, by default None.
    batch_size : int, optional
        Batch size, by default 10.
    num_workers : int, optional
        Number of workers, by default 0.
    **kwargs
        Additional keyword arguments to be passed to the DataLoader.

    Returns
    -------
    DataLoader
        PyTorch DataLoader for the image files. Returns a tuple of (image, label, image_id)
    """
    dataset = SimpleDataset(
        file_paths=file_paths,
        transforms=transforms if transforms is not None else np.array,
    )

    def _custom_collate_fn(batch):
        return zip(*batch)

    return DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=_custom_collate_fn,
        **kwargs,
    )



from abc import ABC, abstractmethod
from typing import Optional

from pymdma.common.output import MetricResult
from pymdma.constants import EvaluationLevel, MetricGroup, ReferenceType


class MetricClass:
    def _post_process_hooks(self, results):
        pass


class InputLayer(ABC):
    def __init__(self):
        self.reference_loader = None
        self.instance_ids = None

    @abstractmethod
    def __len__(self):
        pass

    @property
    def annotations(self):
        raise NotImplementedError

    @property
    def data_properties(self):
        return {}

    @property
    @abstractmethod
    def batched_samples(self):
        pass

    @abstractmethod
    def get_embeddings(self, model_name: str, **kwargs):
        pass


class Metric(ABC):
    # evaluation params
    evaluation_level: EvaluationLevel = EvaluationLevel.DATASET
    metric_group: MetricGroup
    reference_type: ReferenceType = ReferenceType.NONE

    # metric specific
    higher_is_better: Optional[bool] = None
    min_value: Optional[float] = None
    max_value: Optional[float] = None

    def __init__(self, **kwargs) -> None:
        super().__init__()

    @abstractmethod
    def compute(self, *args, **kwargs) -> MetricResult:
        pass


class FeatureMetric(Metric):
    extractor_model_name: str = "default"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


class EmbedderInterface(ABC):
    def __init__(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def extract_features_from_files(self, *args, **kwargs):
        pass

    @abstractmethod
    def _extract_features_dataloader(self, dataloader, **kwargs):
        pass
